import re
import math



class Datavalidation:
    '''
    A class to validate the input macro data, and print errors if any
    
    '''

    def check_variable_names(self, macro_data, macro_data_metadata, source):
        """
        Purpose:
            Validate if the macro data is correct. 
            Checks if macro data contains all variables it is suppose to, and checks if macro data contains unexpected variables
        Inputs:
            macro_data: pandas df with all the macro data
            macro_data_metadata: pandas df with the required metadata to validate data
            source: macro data source
    
        Outputs:
            error_list : list with errors. Will be empty if no errors
    
        """
        error_list = []

        # fetches variable names from the data input and from the metadata file
        variable_names_input_unfiltered = list(macro_data.columns.values)
        variable_names_check_unfiltered = list(macro_data_metadata["Code"])

        # removes '_Actuals', '_Projected', trailing whitespace and tab from input data. Appends to filtered list
        variable_names_input_filtered = []
        for variable in variable_names_input_unfiltered:
            variable = re.sub("\t", "", variable)
            variable = re.sub("_Actuals", "", variable)
            variable = re.sub("_Projections", "", variable)
            variable = variable.rstrip()
            variable_names_input_filtered.append(variable)

        # removes duplicates, and the 'Data' instance(as this is not an actual variable)
        variable_names_input_filtered = list(dict.fromkeys(variable_names_input_filtered))
        variable_names_input_filtered.remove("Date")

        # removes trailing whitespace from metadata variables
        variable_names_check_filtered = []
        for variable in variable_names_check_unfiltered:
            variable = variable.rstrip()
            variable_names_check_filtered.append(variable)

        # loops through metadata variables. Adds error message to error_list if missing in input data
        for variable in variable_names_check_filtered:
            if variable not in variable_names_input_filtered:
                error = variable + " variable missing from " + source
                error_list.append(error)

        # loops through variables in input files. Adds error message if not in metadata file
        for variable in variable_names_input_filtered:
            if variable not in variable_names_check_filtered:
                error = source + " contains variable: " + variable + ". This variable is not in metadata overview. If this is a new variable, please add this to metadata overview"
                error_list.append(error)
        return error_list


    def check_data(self, macro_data, macro_data_metadata, source):
        """
        Purpose:
            Validate if the macro data is correct. 
            Checks if all variables have actuals and projections as specified in macro_data_metadata
            Checks if all macro data is numeric
            Checks if all variables with actuals and projections have a continious time series of actuals and projections
        Inputs:
            macro_data: pandas df with all the macro data
            macro_data_metadata: pandas df with the required metadata to validate data
            source: macro data source
    
        Outputs:
            error_list : list with errors. Will be empty if no errors
    
        """
        error_list = []

        # filters of variables that should have projections
        variables_with_predicted = list(macro_data_metadata.loc[macro_data_metadata["Projections"] != "None"]["Code"])

        # gets list of all variables. Removes Data header
        macro_variables = list(macro_data.columns.values)
        macro_variables.remove("Date")



        # generates list of variables that should have projections
        macro_data_variable_with_projected = []
        for variable in macro_variables:
            compare_variable1 = variable.rstrip()
            compare_variable2 = compare_variable1.replace("_Actuals", "")
            compare_variable3 = compare_variable2.replace("_Projections", "")
            if compare_variable3 in variables_with_predicted:
                macro_data_variable_with_projected.append(variable)


        # iterates over all variables two steps at the time, where i is actuals and i + 1 is projected
        for i in range(0, len(macro_variables), 2):
            actuals_data = list(macro_data[macro_variables[i]])
            projections_data = list(macro_data[macro_variables[i + 1]])

            # checks if variable is in list of variables with projections
            if macro_variables[i] in macro_data_variable_with_projected and macro_variables[i + 1] in macro_data_variable_with_projected:

                # checks if all data is numerical, and sets the boolean variable all_data_is_numerical for avoiding errors later
                all_data_is_numerical = True
                for value in actuals_data:

                    try:
                        float(value)
                    except:
                        try:
                            value = str(value)
                            if " " in value:
                                value = "SPACE"
                        except:
                            pass
                        error = str(macro_variables[i] + " contains nonnumerical value " + value + " in " + source)
                        error_list.append(error)
                        all_data_is_numerical = False

                for value in projections_data:
                    try:
                        float(value)
                    except:
                        try:
                            value = str(value)
                            if " " in value:
                                value = "SPACE"
                        except:
                            pass
                        error = macro_variables[i + 1] + " contains nonnumerical value " + value + " in " + source
                        error_list.append(error)
                        all_data_is_numerical = False

                # to avoid program crashing when checking for math.isnan()
                if all_data_is_numerical:

                    actuals_has_values = False
                    # for locating start of actuals
                    start_counter = 0
                    for data in actuals_data:
                        if not math.isnan(data):
                            start_index = start_counter
                            actuals_has_values = True
                            break
                        start_counter += 1

                    projections_has_values = False
                    # for locating end of projections
                    end_counter = len(projections_data)
                    for data in reversed(projections_data):
                        if not math.isnan(data):
                            end_index = end_counter
                            projections_has_values = True
                            break
                        end_counter -= 1

                    # prints out error if actuals has no data
                    if not actuals_has_values:
                        error = str(macro_variables[i] + " has no values in " + source)
                        error_list.append(error)

                    # prints out error if projections has no data
                    if not projections_has_values:
                        error = str(macro_variables[i + 1] + " has no values in " + source)
                        error_list.append(error)

                    # checks if any data points are missig value between start of actuals and end of projections
                    if actuals_has_values and projections_has_values:
                        for index_number in range(start_index, end_index):
                            if math.isnan(actuals_data[index_number]) and math.isnan(projections_data[index_number]):
                                # +5 is for the excel offset
                                row_with_missing = str(int(index_number + 5))
                                error = str(macro_variables[i] + " or " + macro_variables[i + 1] + " missing data on row " + row_with_missing + " in " + source)
                                error_list.append(error)

            else:

                # for variables without projections, only actuals is controlled for numerical data
                for value in actuals_data:
                    try:
                        float(value)
                    except:
                        try:
                            value = str(value)
                            if " " in value:
                                value = "SPACE"
                        except:
                            pass
                        error = macro_variables[i] + " contains nonnumerical values " + value + " in " + source
                        error_list.append(error)

        return error_list


    def check_macro_input(self, macro_data, macro_data_metadata, source="Markets"):
        """
        Purpose:
            Call functions to check macro data input.
            Checks if variables are forgotten, that there is a continous time series from Actuals to Projections, and that all data is numeric
        Inputs:
            macro_data: pandas df with all the macro data
            macro_metadata: pandas df with the required metadata to validate data
            source: macro data source. When called from Macro Tool project, use Markets or don't pass a value. When called from ECL, use actual macro data source. 
    
        Outputs:
            return_list : list with errors. Will be empty if no errors
    
        """
        error_list = []

        error_list.append(self.check_variable_names(macro_data, macro_data_metadata, source))
        error_list.append(self.check_data(macro_data, macro_data_metadata, source))

        # to transform from list of list to one flat list
        return_list = []
        for sublist in error_list:
            for error in sublist:
                return_list.append(error)

        return return_list

class DataNormalization:
    '''
    A class to normalize the macro data and calculate LTV
    
    '''
    
    def transform_macro_data(self, macro_data, macro_metadata):
        """
        Purpose:
            Transform macro data in accordance with the information taken from macro_metadata.
            This will ensure all macro data are on the same value format
    
        Inputs:
            macro_data: pandas df with all the macro data
            macro_metadata: pandas df with the required metadata to do normalization
    
        Outputs:
            macro_data : the same dataframe as in the input, but with actuals and projetions normalized
    
        """

        # use macro_metadata to find variables where projected must be transformed
        variables_to_be_transformed = []
        for index, variable in macro_metadata.iterrows():
            if variable[2] == "Value" and variable[3] == "%":
                variables_to_be_transformed.append(variable[0])

        # iterates through all the variables that must be transformed
        for transform_variable in variables_to_be_transformed:

            # iterate through all the rows of actuals and projections
            for i in range(len(macro_data[transform_variable + "_Actuals"])):

                # checks if actuals is nan and projected has a value. If yes, that projected value must be transformed
                if math.isnan(macro_data.loc[i][transform_variable + "_Actuals"]) and not math.isnan(macro_data.loc[i][transform_variable + "_Projections"]):

                    # checks if the previous row of actuals is nan. If not, this value is used to transformed projected
                    if not math.isnan(macro_data.loc[i - 1][transform_variable + "_Actuals"]):
                        last_value = macro_data.loc[i - 1][transform_variable + "_Actuals"]

                    # if previous row of actuals is nan, the previous row of projections is used for the transformation
                    else:
                        last_value = macro_data.loc[i - 1][transform_variable + "_Projections"]

                    # projections data is transformed by last value * (1 + (projections value / 100))
                    macro_data.loc[i, transform_variable + "_Projections"] = ((macro_data.loc[i][transform_variable + "_Projections"]/100) + 1) * last_value

        # use macro_metadata to find variables where actuals and projected must be transformed
        variables_to_be_transformed.clear()
        for index, variable in macro_metadata.iterrows():
            if variable[2] == "%" and variable[3] == "%":
                variables_to_be_transformed.append(variable[0])

        # iterates through all the variables where _actuals and _projections must be performed
        for transform_variable in variables_to_be_transformed:

            # iterate through all the rows of actuals and projections
            for i in range(len(macro_data[transform_variable + "_Actuals"])):

                #checks if value is nan, if not it is transformed to percent
                if not math.isnan(macro_data.loc[i][transform_variable + "_Actuals"]):
                    macro_data.loc[i, transform_variable + "_Actuals"] = float(macro_data.loc[i, transform_variable + "_Actuals"] /100)

                if not math.isnan(macro_data.loc[i][transform_variable + "_Projections"]):
                    macro_data.loc[i, transform_variable + "_Projections"] = float(macro_data.loc[i, transform_variable + "_Projections"] /100)

        return macro_data


    def generate_LTV_hh(self, dataframe, index_years):
        """
        Purpose:
            Generate the time series for Loan to value(LTV). A parameter possible to used in some macro models, but where we lack projections
    
        Inputs:
            dataframe: pandas dataframe with macro values
            index_years: dataframe with index years for HPI_index and c2_hh_index, specifying what divider to use
    
        Outputs:
            dataframe : the same dataframe as in the input, but with actuals and projetions for LTV
    
        """

        dataframe.set_index("Date", inplace=True)
        
        # gets required time series from macro data dataframe
        C2_hh_Actuals = list(dataframe['C2_hh_Actuals'])
        C2_hh_Projections = list(dataframe['C2_hh_Projections'])
        HPI_Actuals = list(dataframe['HPI_Actuals'])
        HPI_Projections = list(dataframe['HPI_Projections'])
        
        # gets HPI divider
        HPI_index_year_zero = index_years.loc["HPI_index"]["Date"].date()
        HPI_divider = dataframe.loc[HPI_index_year_zero]['HPI_Actuals']
        
        # gets C2_hh divider
        C2_hh_index_year_zero = index_years.loc["c2_hh_index"]["Date"]
        C2_hh_divider = dataframe.loc[C2_hh_index_year_zero]['C2_hh_Actuals']
        
        # calculate LTV actuals
        LTV_hh_Actuals = []
        for i in range(len(C2_hh_Actuals)):
            if math.isnan(C2_hh_Actuals[i]) or math.isnan(HPI_Actuals[i]):
                LTV_hh_Actuals.append(None)
            else:
                LTV_hh_Actuals.append((C2_hh_Actuals[i]/C2_hh_divider)/(HPI_Actuals[i]/HPI_divider))
        
        # saves results to macro dataframe
        dataframe['LTV_hh_Actuals'] = LTV_hh_Actuals
        
        # calculate LTV projections
        LTV_hh_Projections = []
        for i in range(len(C2_hh_Projections)):
            if math.isnan(C2_hh_Projections[i]) or math.isnan(HPI_Projections[i]):
                LTV_hh_Projections.append(None)
            else:
                LTV_hh_Projections.append((C2_hh_Projections[i]/C2_hh_divider) / (HPI_Projections[i]/HPI_divider))
        
        # saves results to macro dataframe
        dataframe['LTV_hh_Projections'] = LTV_hh_Projections
        
        dataframe.reset_index(inplace=True)
        return dataframe

class Ols_parameter_transformation:

    def transform_variable(self, actuals, transform_type, projections=None, US_actual_inflation=None, US_projected_inflation=None,
                           NOR_actual_inflation=None, NOR_projected_inflation=None, inflation_metadata=None):
        """
            Purpose:
                Calls the functions corresponding with the type input.
                Inflation adjusting the variables and transforming to natural logarithm is done in separate if-statements.
                Other transformations are stored as lambda functions in transform_functions_dict.

            Inputs:
                actuals: list with tuples of date and actual time series to be transformed
                transform_type: the transformation to be done
                projections: list with tuples of date and projected time series to be transformed
                US_actual_inflation: list of tuples with date and actual US inflation time series
                US_projected_inflation: list of tuples with date and projected US inflation time series
                NOR_actual_inflation: list of tuples with date and actual Norwegian inflation time series
                NOR_projected_inflation: list of tuples with date and projected Norwegian inflation time series
                inflation_metadata: pandas df with inflation metadata to know what is considered year zero for Norwegian and US inflation

            Outputs:
                return_dict: dictionary with key = dates and value = transformed value per observation date

            """

        input_dict = {}
        
        # creates one continuous time series of actuals and projections
        for i in range(len(actuals[0])):
            if not math.isnan(actuals[1][i]):
                input_dict[actuals[0][i]] = (actuals[1][i], "_Actuals")
            elif not math.isnan(projections[1][i]):
                input_dict[actuals[0][i]] = (projections[1][i], "_Projections")

        # if type is not string, it skips right to return
        if isinstance(transform_type, str):

            # dictionary that maps transformation name to functionality
            transform_functions_dict = \
               {

                "Lag1": lambda: self.transform_lag(input_dict=input_dict, year_input=1),
                "Lag2": lambda: self.transform_lag(input_dict=input_dict, year_input=2),
                "Lag3": lambda: self.transform_lag(input_dict=input_dict, year_input=3),

                "Pct1": lambda: self.transform_pct_change(input_dict=input_dict, year_input=1),
                "Pct2": lambda: self.transform_pct_change(input_dict=input_dict, year_input=2),
                "Pct3": lambda: self.transform_pct_change(input_dict=input_dict, year_input=3),

                "Delta1": lambda: self.transform_delta(input_dict=input_dict, year_input=1),
                "Delta2": lambda: self.transform_delta(input_dict=input_dict, year_input=2),
                "Delta3": lambda: self.transform_delta(input_dict=input_dict, year_input=3),
                "Delta4": lambda: self.transform_delta(input_dict=input_dict, year_input=4),

                "Maxdelta2": lambda: self.transform_max_delta(input_dict=input_dict, year_input=2),
                "Maxdelta3": lambda: self.transform_max_delta(input_dict=input_dict, year_input=3),
                "Maxdelta4": lambda: self.transform_max_delta(input_dict=input_dict, year_input=4),
                
                "Pct1Maxdelta2": lambda: self.transform_max_delta(input_dict=self.transform_pct_change(input_dict=input_dict, year_input=1), year_input=2),
                "Pct1Maxdelta3": lambda: self.transform_max_delta(input_dict=self.transform_pct_change(input_dict=input_dict, year_input=1), year_input=3),
                "Pct1MMaxdelta4": lambda: self.transform_max_delta(input_dict=self.transform_pct_change(input_dict=input_dict, year_input=1), year_input=4),

                "Movavg2": lambda: self.transform_moving_avg(input_dict=input_dict, year_input=2),
                "Movavg3": lambda: self.transform_moving_avg(input_dict=input_dict, year_input=3),
                "Movavg4": lambda: self.transform_moving_avg(input_dict=input_dict, year_input=4),

                "Pct1Movavg2": lambda: self.transform_moving_avg(input_dict=self.transform_pct_change(input_dict=input_dict, year_input=1), year_input=2),
                "Pct1Movavg3": lambda: self.transform_moving_avg(input_dict=self.transform_pct_change(input_dict=input_dict, year_input=1), year_input=3),
                "Pct1Movavg4": lambda: self.transform_moving_avg(input_dict=self.transform_pct_change(input_dict=input_dict, year_input=1), year_input=4),

                "Lag1Pct1": lambda: self.transform_lag(input_dict=self.transform_pct_change(input_dict=input_dict, year_input=1), year_input=1),
                "Lag2Pct1": lambda: self.transform_lag(input_dict=self.transform_pct_change(input_dict=input_dict, year_input=1), year_input=2),
                "Lag3Pct1": lambda: self.transform_lag(input_dict=self.transform_pct_change(input_dict=input_dict, year_input=1), year_input=3),
                "Lag1Pct2": lambda: self.transform_lag(input_dict=self.transform_pct_change(input_dict=input_dict, year_input=2), year_input=1),
                "Lag2Pct2": lambda: self.transform_lag(input_dict=self.transform_pct_change(input_dict=input_dict, year_input=2), year_input=2),
                "Lag3Pct2": lambda: self.transform_lag(input_dict=self.transform_pct_change(input_dict=input_dict, year_input=2), year_input=3),
                "Lag1Pct3": lambda: self.transform_lag(input_dict=self.transform_pct_change(input_dict=input_dict, year_input=3), year_input=1),
                "Lag2Pct3": lambda: self.transform_lag(input_dict=self.transform_pct_change(input_dict=input_dict, year_input=3), year_input=2),
                "Lag3Pct3": lambda: self.transform_lag(input_dict=self.transform_pct_change(input_dict=input_dict, year_input=3), year_input=3),

                "Lag1Delta1": lambda: self.transform_lag(input_dict=self.transform_delta(input_dict=input_dict, year_input=1), year_input=1),
                "Lag2Delta1": lambda: self.transform_lag(input_dict=self.transform_delta(input_dict=input_dict, year_input=1), year_input=2),
                "Lag3Delta1": lambda: self.transform_lag(input_dict=self.transform_delta(input_dict=input_dict, year_input=1), year_input=3),
                "Lag1Delta2": lambda: self.transform_lag(input_dict=self.transform_delta(input_dict=input_dict, year_input=2), year_input=1),
                "Lag2Delta2": lambda: self.transform_lag(input_dict=self.transform_delta(input_dict=input_dict, year_input=2), year_input=2),
                "Lag3Delta2": lambda: self.transform_lag(input_dict=self.transform_delta(input_dict=input_dict, year_input=2), year_input=3),
                "Lag1Delta3": lambda: self.transform_lag(input_dict=self.transform_delta(input_dict=input_dict, year_input=3), year_input=1),
                "Lag2Delta3": lambda: self.transform_lag(input_dict=self.transform_delta(input_dict=input_dict, year_input=3), year_input=2),
                "Lag3Delta3": lambda: self.transform_lag(input_dict=self.transform_delta(input_dict=input_dict, year_input=3), year_input=3),

                "Lag1Maxdelta2": lambda: self.transform_lag(input_dict=self.transform_max_delta(input_dict=input_dict, year_input=2), year_input=1),
                "Lag2Maxdelta2": lambda: self.transform_lag(input_dict=self.transform_max_delta(input_dict=input_dict, year_input=2), year_input=2),
                "Lag3Maxdelta2": lambda: self.transform_lag(input_dict=self.transform_max_delta(input_dict=input_dict, year_input=2), year_input=3),
                "Lag1Maxdelta3": lambda: self.transform_lag(input_dict=self.transform_max_delta(input_dict=input_dict, year_input=3), year_input=1),
                "Lag2Maxdelta3": lambda: self.transform_lag(input_dict=self.transform_max_delta(input_dict=input_dict, year_input=3), year_input=2),
                "Lag3Maxdelta3": lambda: self.transform_lag(input_dict=self.transform_max_delta(input_dict=input_dict, year_input=3), year_input=3),
                "Lag1Maxdelta4": lambda: self.transform_lag(input_dict=self.transform_max_delta(input_dict=input_dict, year_input=4), year_input=1),
                "Lag2Maxdelta4": lambda: self.transform_lag(input_dict=self.transform_max_delta(input_dict=input_dict, year_input=4), year_input=2),
                "Lag3Maxdelta4": lambda: self.transform_lag(input_dict=self.transform_max_delta(input_dict=input_dict, year_input=4), year_input=3),

                "Lag1Movavg2": lambda: self.transform_lag(input_dict=self.transform_moving_avg(input_dict=input_dict, year_input=2), year_input=1),
                "Lag2Movavg2": lambda: self.transform_lag(input_dict=self.transform_moving_avg(input_dict=input_dict, year_input=2), year_input=2),
                "Lag3Movavg2": lambda: self.transform_lag(input_dict=self.transform_moving_avg(input_dict=input_dict, year_input=2), year_input=3),
                "Lag1Movavg3": lambda: self.transform_lag(input_dict=self.transform_moving_avg(input_dict=input_dict, year_input=3), year_input=1),
                "Lag2Movavg3": lambda: self.transform_lag(input_dict=self.transform_moving_avg(input_dict=input_dict, year_input=3), year_input=2),
                "Lag3Movavg3": lambda: self.transform_lag(input_dict=self.transform_moving_avg(input_dict=input_dict, year_input=3), year_input=3),
                "Lag1Movavg4": lambda: self.transform_lag(input_dict=self.transform_moving_avg(input_dict=input_dict, year_input=4), year_input=1),
                "Lag2Movavg4": lambda: self.transform_lag(input_dict=self.transform_moving_avg(input_dict=input_dict, year_input=4), year_input=2),
                "Lag3Movavg4": lambda: self.transform_lag(input_dict=self.transform_moving_avg(input_dict=input_dict, year_input=4), year_input=3),
                 }

            # checks if Real is part of transformation type
            if "Real" in transform_type:

                if "RealUS" in transform_type:
                    actual_inflation = US_actual_inflation
                    projected_inflation = US_projected_inflation

                    inflation_year_zero = inflation_metadata.loc["USINF"]["Date"]

                    # removes substring from transformation type
                    transform_type = transform_type.replace("RealUS", "")

                else:
                    actual_inflation = NOR_actual_inflation
                    projected_inflation = NOR_projected_inflation

                    inflation_year_zero = inflation_metadata.loc["NOINF"]["Date"]

                    # removes substring from transformation type
                    transform_type = transform_type.replace("RealNOR", "")

                inflation_dict = {}
                
                # creates one continuous time series of actuals and projections
                for i in range(len(actual_inflation[0])):
                    if not math.isnan(actual_inflation[1][i]):
                        inflation_dict[actual_inflation[0][i]] = (actual_inflation[1][i], "_Actuals")
                    elif not math.isnan(projected_inflation[1][i]):
                        inflation_dict[actual_inflation[0][i]] = (projected_inflation[1][i], "_Projections")

                
                # calls function to inflation adjust input dict
                input_dict = self.transform_inflation_adjust(input_dict=input_dict, inflation_dict=inflation_dict, inflation_year_zero=inflation_year_zero)

            # checks if Ln is part of transformation type
            if "Ln" in transform_type:
                #calls function to transform input_dict to ln
                input_dict = self.transform_natural_logarithm(input_dict=input_dict)

                #removes Ln from transformation type
                transform_type = transform_type.replace("Ln", "")

            # checks if transformation needs to be done
            if transform_type != None and transform_type != "":

                # aborts execution if not transformation is in transform_functions_dict
                assert transform_type in transform_functions_dict, str("Type of transformation " + transform_type +
                " is not supported. Add transformation type to transform_function_dict in function transform_Variable in script Dataprep_Functions.py")
                # calls the function from transform_functions_dict the corresponds with Type
                return_dict = transform_functions_dict.get(transform_type)()
            else:
                return_dict = input_dict

        else:
            # if no transformaton is to be done, return_dict = input_dict
            return_dict = input_dict

        return return_dict


    def transform_lag(self, input_dict, year_input):  
        """
        Purpose:
            Transforms the macro variables of given date to value of given date - X years
            X is specified by the input year_input
    
        Inputs:
            input_dict: dictionary with key = date, value = set(macro variable value, actuals/projections)
            year_input: an integer specifying how many years back to find value to find delta with
    
        Outputs:
            return dict : dictionary with key = date, value = set(transformed macro variable, actuals/projections)
    
        """
        return_dict = {}
        
        # add prediction points equal to lag years
        max_year = max(input_dict.keys())
        for lag_year in range(1, year_input + 1):
            new_date = max_year.replace(year=max_year.year + lag_year)
            input_dict[new_date] = (0, "_Projections")

        for row in input_dict:
            # checks if the date of the row - year_delta is in input_dict.
            if row.replace(year=row.year - year_input) in input_dict:
                # sets row date equal to value of row date - year_delta.
                return_dict[row] = (input_dict[row.replace(year=row.year - year_input)][0], input_dict[row][1])

        return return_dict



    def transform_pct_change(self, input_dict, year_input):
        """
        Purpose:
            Transforms the macro variables to the delta in percent between given year and the value of given year - X years
            X is specified by the input year_input
    
        Inputs:
            input_dict: dictionary with key = date, value = set(macro variable value, actuals/projections)
            year_input: an integer specifying how many years back to find value to find delta percent with
    
        Outputs:
            return dict : dictionary with key = date, value = set(transformed macro variable, actuals/projections)
    
        """
        return_dict = {}
        
        # calculates percentage change from year to year
        for row in input_dict:
            if row.replace(year=row.year - year_input) in input_dict:
                return_dict[row] = (input_dict[row][0]/input_dict[row.replace(year=row.year - year_input)][0] - 1, input_dict[row][1])

        return return_dict



    def transform_delta(self, input_dict, year_input):
        """
        Purpose:
            Transforms the macro variables to the delta between given year and the value of given year - X
            X is specified by the input year_input
    
        Inputs:
            input_dict: dictionary with key = date, value = set(macro variable value, actuals/projections)
            year_input: an integer specifying how many years back to find value to find delta with
    
        Outputs:
            return dict : dictionary with key = date, value = set(transformed macro variable, actuals/projections)
    
        """
        return_dict = {}
        
        # calculates change from year to year
        for row in input_dict:
            if row.replace(year=row.year - year_input) in input_dict:
                return_dict[row] = (input_dict[row][0] - input_dict[row.replace(year=row.year - year_input)][0], input_dict[row][1])

        return return_dict



    def transform_moving_avg(self, input_dict, year_input):
        """
        Purpose:
            Transforms macro variable to the moving average of the last X years. 
            X is given by the input year_input
    
        Inputs:
            input_dict: dictionary with key = date, value = set(macro variable value, actuals/projections)
            year_input: an integer specifying how many years back to find the average of
    
        Outputs:
            return dict : dictionary with key = date, value = set(transformed macro variable, actuals/projections)
    
        """
        return_dict = {}
        
        # calculates the moving average for x number of years
        for row in input_dict:
            if row.replace(year=row.year - (year_input - 1)) in input_dict:
                sum = 0
                for i in range(year_input):
                    sum += input_dict[row.replace(year=row.year - i)][0]
                average = sum/year_input
                return_dict[row] = (average, input_dict[row][1])
        return return_dict


    def transform_natural_logarithm(self, input_dict):
        """
        Purpose:
            Transform macro variables to natural logarithm
    
        Inputs:
            input_dict: dictionary with key = date, value = set(macro variable value, actuals/projections)
    
        Outputs:
            return dict : dictionary with key = date, value = set(natural logarithm transformed macro variable, actuals/projections)
    
        """
        return_dict = {}
        
        # calculates the natural logarithm
        for row in input_dict:
            return_dict[row] = (math.log(input_dict[row][0]), input_dict[row][1])
        return return_dict



    def transform_max_delta(self, input_dict, year_input):
        """
        Purpose:
            Transform macro data to the delta between value of given year and the maximum value of the X previous years. 
            X is given by the input_variable year_input. 
    
        Inputs:
            input_dict: dictionary with key = date, value = set(macro variable value, actuals/projections)
            year_input: integer input specifying how many years back to find max value used to find max delta
    
        Outputs:
            return dict : dictionary with key = date, value = set(transformed macro variable, actuals/projections)
    
        """
        return_dict = {}
        
        # calculates the max delta
        for row in input_dict:
            if row.replace(year=row.year - year_input) in input_dict:
                max_list = []
                for i in range(1, year_input + 1):
                    max_list.append(input_dict[row.replace(year=row.year - i)][0])

                max_value = max(max_list)
                return_dict[row] = (input_dict[row][0] - max_value, input_dict[row][1])

        return return_dict



    def transform_inflation_adjust(self, input_dict, inflation_dict, inflation_year_zero):
        """
        Purpose:
            Inflation adjusts the macro variables
    
        Inputs:
            input_dict: dictionary with key = date, value = set(macro variable value, actuals/projections)
            inflation_dict: dictionary with key = date, value = set(inflation value, actuals/projections)
    
        Outputs:
            return dict : dictionary with key = date, value = set(inflation adjusted macro variable, actuals/projections)
    
        """
        return_dict = {}
        
        # calculate inflation adjusted values
        for row in input_dict:
            if row in inflation_dict:
                return_dict[row] = (input_dict[row][0] / (inflation_dict[row][0] / inflation_dict[inflation_year_zero][0]), input_dict[row][1])
        return return_dict
