import pandas as pd
import statsmodels.api as sm
import numpy as np
import itertools
import multiprocessing as mp
import time
import statistics
from functools import partial
from scipy.stats import norm
import shutil
import openpyxl
import os

# imports .py file in different folder
import sys
sys.path.append("../Dataprep")
import Dataprep_Functions

# ignores warnings from the program
import warnings
warnings.simplefilter("ignore")


def import_and_clean_macro_input(path_dict, macro_type):
    """
               Purpose:
               Imports macro data, validates it and normalize it

               Inputs:
               Path_dict: dictionary of paths to required .xlsx-files

               Outputs:
               Dictionary with macro values{dnb_markets_macro_transformed, us_inflation_actuals, us_inflation_projections, nor_inflation_actuals, nor_inflation_projections, dates, index_years}

    """
    # reads index years from macro metadata
    index_years = pd.read_excel(path_dict["Macro_metadata"], sheet_name="IndexYears")
    index_years.set_index("Code", inplace=True)
    
    # reads in the macro data
    sheet_name = sheet_name=path_dict["Macro_sheet_dict"][macro_type]
    dnb_markets_macro = pd.read_excel(path_dict["Macro_variables"], sheet_name,
                                      header=3)
    # strips the column names of leading and trailing whitespace
    dnb_markets_macro.rename(columns=lambda x: x.strip(), inplace=True)
    
    # reads in the macro data for use in validation
    dnb_markets_macro_metadata = pd.read_excel(path_dict["Macro_metadata"], sheet_name="VariableTypeMarkets")
    
    # calls functionality to do validation of macro input
    input_validation = Dataprep_Functions.Datavalidation()
    error_list = input_validation.check_macro_input(dnb_markets_macro, dnb_markets_macro_metadata, source="DNB Markets data")
    
    # transfer the errors to a list to be printed for user
    no_errors = True
    error_output_list = []
    for error in error_list:
        if error:
            error_output_list.append(error)
            no_errors = False
    
    
    # prints out the errors in the macro input
    if not no_errors:
        print("The following errors were found in the macro data input for", macro_type, ":")
        
        for error in error_output_list:
            print(error)
    
    # gives the user the option to continue despite errors
    # it was decided that this was unwise to allow for in the current program, but might be changed later when the macro input validation is improved
    '''
    if not no_errors:
        print("Running the program with these errors might cause the program to malfunction")
        user_error_reply = input("If you want to abort program press enter. If your still want to run the program, type 'gogo' and hit enter:")

        if user_error_reply.lower() == "gogo" or user_error_reply.lower() == "'gogo'":
            print("Resuming program")
            no_errors = True

        else:
            print("Aborting program")
            no_errors = False
    '''   
    
    # if no errors, the data is normalized, and the data from this is extracted to variables
    if no_errors:

        data_normalization = Dataprep_Functions.DataNormalization()
        dnb_markets_macro_transformed = data_normalization.transform_macro_data(dnb_markets_macro,                                                                                dnb_markets_macro_metadata)
        dnb_markets_macro_transformed = data_normalization.generate_LTV_hh(dnb_markets_macro_transformed, index_years)
        
        dates = dnb_markets_macro_transformed["Date"]

        us_inflation_actuals = dnb_markets_macro_transformed["USINF_Actuals"]
        us_inflation_actuals = (dates, us_inflation_actuals)

        us_inflation_projections = dnb_markets_macro_transformed["USINF_Projections"]
        us_inflation_projections = (dates, us_inflation_projections)

        nor_inflation_actuals = dnb_markets_macro_transformed["NOINF_Actuals"]
        nor_inflation_actuals = (dates, nor_inflation_actuals)

        nor_inflation_projections = dnb_markets_macro_transformed["NOINF_Projections"]
        nor_inflation_projections = (dates, nor_inflation_projections)
        
        return_dict = {"dnb_markets_macro_transformed": dnb_markets_macro_transformed, "us_inflation_actuals": us_inflation_actuals, "us_inflation_projections": us_inflation_projections, "nor_inflation_actuals": nor_inflation_actuals, "nor_inflation_projections": nor_inflation_projections, "dates": dates, "index_years": index_years}
        
        return return_dict
        
    else:
        print("Please correct these errors in the input macro file, and upload it again")
        return False
    
    
def get_Z_values(path_dict, model_name):
    """
               Purpose:
               Gets the Z value and observation dates to be modelled. 

               Outputs: a dictionary with a list of z values and lsit of corresponding dates                     

    """
    
    z_credit_cycle = pd.read_excel(path_dict["Z_values"], sheet_name=model_name)

    z_date_list = list(z_credit_cycle["Date"])
    z_values = list(z_credit_cycle["Z"])
    
    return {"z_date_list":z_date_list, "z_values":z_values}
    
    
def in_depth_analysis(macro_model_name, explanatory_variables, path_dict, rolling_years_amount=16):
    """
               Purpose:
               To give the user required information ion order to pick the optimal OLS regression model

               Inputs:
                macro_model_name: string with name of macto model to estimate
                explanatory_variables: list with tuples containing name of explanatory variable and transformation
                path_dict: dictionary with the path to the following files: macro variable data from markets, stressed macro variables,
                low scenario macro variables, macro meta data, estimated Z, 
                rolling_years_amount: integer specifying how many years to do rolling regression for

               Outputs:
                RMSE comparison of chosen model to as-is and TTC. Coefficients from rolling regression.
                R-squared and MSE of model, p-values and coefficients of variables
                KPSS statistics
                Predictions for baseline, stressed and low macro scenarions                

    """
    # calls function to get Z time series to be modelled
    z_credit_cycle = get_Z_values(path_dict, macro_model_name)    
    z_date_list = z_credit_cycle["z_date_list"]
    z_values = z_credit_cycle["z_values"]

    # imports, validates and cleans macro data
    macro_data_dict = import_and_clean_macro_input(path_dict, "Baseline_sheet")
    adverse_macro_data_dict = import_and_clean_macro_input(path_dict, "Adverse_sheet")
    low_macro_data_dict = import_and_clean_macro_input(path_dict, "Low_sheet")
    
    # extract macro data from dictionaries
    if macro_data_dict and adverse_macro_data_dict and low_macro_data_dict:
        dnb_markets_macro_transformed = macro_data_dict["dnb_markets_macro_transformed"]
        us_inflation_actuals = macro_data_dict["us_inflation_actuals"]
        us_inflation_projections = macro_data_dict["us_inflation_projections"]
        nor_inflation_actuals = macro_data_dict["nor_inflation_actuals"]
        nor_inflation_projections = macro_data_dict["nor_inflation_projections"]
        dates = macro_data_dict["dates"]
        index_years = macro_data_dict["index_years"]
        
        adverse_dnb_markets_macro_transformed = adverse_macro_data_dict["dnb_markets_macro_transformed"]
        adverse_us_inflation_actuals = adverse_macro_data_dict["us_inflation_actuals"]
        adverse_us_inflation_projections = adverse_macro_data_dict["us_inflation_projections"]
        adverse_nor_inflation_actuals = adverse_macro_data_dict["nor_inflation_actuals"]
        adverse_nor_inflation_projections = adverse_macro_data_dict["nor_inflation_projections"]
        adverse_dates = adverse_macro_data_dict["dates"]
        
        low_dnb_markets_macro_transformed = low_macro_data_dict['dnb_markets_macro_transformed']
        low_us_inflation_actuals = low_macro_data_dict["us_inflation_actuals"]
        low_us_inflation_projections = low_macro_data_dict["us_inflation_projections"]
        low_nor_inflation_actuals = low_macro_data_dict["nor_inflation_actuals"]
        low_nor_inflation_projections = low_macro_data_dict["nor_inflation_projections"]
        low_dates = low_macro_data_dict["dates"]
        
        # generate dict of all possible dummies for the given time series by calling custom function generate_dummies()
        dummy_dict = generate_dummies(z_date_list)
        
        transformation_object = Dataprep_Functions.Ols_parameter_transformation()
        variable_time_series = {}        
        dummy_time_series = {}
        
        adverse_variable_time_series = {}
        low_variable_time_series = {}
        
        # iterate through the variables and do transformations on the data
        for variable in explanatory_variables:
            
            # dummies will not be transformed
            if "Dummy" not in variable[0]:
                
                variable_name = variable[0]
                transformation = variable[1]
                
                # transform actuals and projections for baseline, low and adverse scenarios, and transform value to floats
                
                macro_data_actuals = dnb_markets_macro_transformed[variable_name +"_Actuals"]
                macro_data_actuals = pd.Series(list(map(lambda x: float(x), macro_data_actuals)))

                macro_data_projections = dnb_markets_macro_transformed[variable_name + "_Projections"]
                macro_data_projections = pd.Series(list(map(lambda x: float(x), macro_data_projections)))
                
                adverse_macro_data_actuals = adverse_dnb_markets_macro_transformed[variable_name +"_Actuals"]
                adverse_macro_data_actuals = pd.Series(list(map(lambda x: float(x), adverse_macro_data_actuals)))
                
                adverse_macro_data_projections = adverse_dnb_markets_macro_transformed[variable_name + "_Projections"]
                adverse_macro_data_projections = pd.Series(list(map(lambda x: float(x), adverse_macro_data_projections)))
                
                low_macro_data_actuals = low_dnb_markets_macro_transformed[variable_name +"_Actuals"]
                low_macro_data_actuals = pd.Series(list(map(lambda x: float(x), low_macro_data_actuals)))
                
                low_macro_data_projections = low_dnb_markets_macro_transformed[variable_name + "_Projections"]
                low_macro_data_projections = pd.Series(list(map(lambda x: float(x), low_macro_data_projections)))
                
                # creates lists of tuples with dates and macro data values. This format is required by the program
                macro_data_list_actuals = (dates, macro_data_actuals)
                macro_data_list_projections = (dates, macro_data_projections)
                
                adverse_macro_data_list_actuals = (adverse_dates, adverse_macro_data_actuals)
                adverse_macro_data_list_projections = (adverse_dates, adverse_macro_data_projections)
                
                low_macro_data_list_actuals = (low_dates, low_macro_data_actuals)
                low_macro_data_list_projections = (low_dates, low_macro_data_projections)
                
                
                # calls function to transform variable
                variable_time_series[(variable)] = transformation_object.transform_variable(actuals=macro_data_list_actuals, transform_type=transformation, projections=macro_data_list_projections, US_actual_inflation=us_inflation_actuals, US_projected_inflation=us_inflation_projections, NOR_actual_inflation=nor_inflation_actuals, NOR_projected_inflation=nor_inflation_projections, inflation_metadata=index_years)
                
                # calls function to transform adverse variables
                adverse_variable_time_series[(variable)] = transformation_object.transform_variable(actuals=adverse_macro_data_list_actuals, transform_type=transformation, projections=adverse_macro_data_list_projections, US_actual_inflation=adverse_us_inflation_actuals, US_projected_inflation=adverse_us_inflation_projections, NOR_actual_inflation=adverse_nor_inflation_actuals, NOR_projected_inflation=adverse_nor_inflation_projections, inflation_metadata=index_years)
                
                # calls function to transform low variables
                low_variable_time_series[(variable)] = transformation_object.transform_variable(actuals=low_macro_data_list_actuals, transform_type=transformation, projections=low_macro_data_list_projections, US_actual_inflation=low_us_inflation_actuals, US_projected_inflation=low_us_inflation_projections, NOR_actual_inflation=low_nor_inflation_actuals, NOR_projected_inflation=low_nor_inflation_projections, inflation_metadata=index_years)
                
            else:
                # gets the dummy time series corresponding with the given dummy name
                dummy_time_series[(variable[0])] = dummy_dict[variable[0]]
        
        # calls function to trim the time series to be same length
        OLS_input = trim_time_series(z_values, variable_time_series, dummy_time_series, z_date_list, adverse_variable_time_series, low_variable_time_series)
        
        # calls function to do OLS regression and get extended statistical info. Returns an object of class OLS_result_object
        model_results = do_extended_OLS(z_time_series=OLS_input[0], variable_time_series_list=OLS_input[1], variable_names=OLS_input[2], baseline_predictions=OLS_input[4], adverse_predictions=OLS_input[5], low_predictions=OLS_input[6])
        
        # calls function to do rolling regression. Returns a dict with results
        rolling_regression_dict = rolling_regression(z_time_series=OLS_input[0], variable_time_series_list=OLS_input[1], years_to_roll=rolling_years_amount)
        
        # appends the results to object of class OLS_result_object
        model_results.set_rolling_regression_statistics(rolling_regression_dict)
        
        # converts OLS timestamps to dates and appends to model results
        z_dates = OLS_input[3]
        z_dates = list(map(lambda x: x.date(), z_dates))
        model_results.set_dates(z_dates)
        
        # sets model name and variables
        model_results.set_model_name(macro_model_name)
        model_results.set_variables(explanatory_variables)
        
        # calls function to write out results to excel
        write_in_depth_results_to_excel(model_results, path_dict)
        
        return model_results
    
    # returns false if program stopped by macro input validation 
    else:
        return False
        

def write_in_depth_results_to_excel(model_results, path_dict):
    """
               Purpose:
               Writing out the results to a structured excel file

               Inputs:
                model_results: an object of class OLS_result_object with the required data
                path_dict: dictionary containing path to result template and to output directory

               Outputs:
                An excel file based on the excel file used as a template with required statistical info               

    """
    # gets timestamp to create a unique name of output file
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    
    # gets path to template for output and where to save output
    template_path = path_dict["OLS_output_template"]
    OLS_results_ouput_path = path_dict["OLS_output_path"]
    
    # creates output file name from variable names and timestamp
    variable_string = "_".join(list(map(lambda x: x[0], model_results.get_variables())))    
    OLS_results_output_name = str(model_results.get_model_name() + "/" + variable_string + "_" + timestamp + ".xlsx")
    
    # checks if this macro model has directory for output. If not, a directory is created
    model_output_directory = []
    for file in os.listdir(OLS_results_ouput_path):
        model_output_directory.append(file)
        
    if model_results.get_model_name() not in model_output_directory:
        os.mkdir(str(OLS_results_ouput_path + model_results.get_model_name()))
    
    # copies template and change name and directory location
    shutil.copy(template_path, str(OLS_results_ouput_path + OLS_results_output_name))
    
    # loads workbook and sheets
    myworkbook=openpyxl.load_workbook(str(OLS_results_ouput_path + OLS_results_output_name))
    worksheet_data= myworkbook.get_sheet_by_name('Summary graph data')
    worksheet_rr = myworkbook.get_sheet_by_name('Rolling regression ext. info')
    worksheet_summary = myworkbook.get_sheet_by_name('Summary')
    worksheet_variables = myworkbook.get_sheet_by_name('Variables extended info')
    worksheet_time_series = myworkbook.get_sheet_by_name('Time series')
    
    # gets and inserts data to update the graph comparing predicted and actual Z
    z_dates = model_results.get_dates()
    Z = model_results.get_z()
    pred = model_results.get_predictions()
    prediction_intervals = model_results.get_confidence_intervals()
    
    # gets predictions for the different scenarios
    baseline_prediction = model_results.get_baseline_predictions()
    adverse_prediction = model_results.get_adverse_predictions()
    low_prediction = model_results.get_low_predictions()

    # inserts values of actual and predicted Z for the range used for modelling, and dates for the range
    for i in range(len(z_dates)):
        worksheet_data.cell(row=(i+2), column=1).value = z_dates[i]

    for i in range(len(Z)):
        worksheet_data.cell(row=(i+2), column=2).value = Z[i]

    # sets predicted Z for all scenarios equal to baseline predictions for the modelled periode(before future projections), as this is based on historical values that are identical
    for i in range(len(pred)):
        worksheet_data.cell(row=(i+2), column=3).value = pred[i]
        worksheet_data.cell(row=(i+2), column=6).value = pred[i]
        worksheet_data.cell(row=(i+2), column=7).value = pred[i]
        worksheet_data.cell(row=(i+2), column=4).value = prediction_intervals["Lower"][i]
        worksheet_data.cell(row=(i+2), column=5).value = prediction_intervals["Upper"][i]
    
    # in case prediction length differs for baseline, adverse and low. Gets the longest prediction
    prediction_length = max(list(map(lambda x: len(x), [baseline_prediction, adverse_prediction, low_prediction])))
    
    # prints Z predictions for each scenario. Use try-except as time series length might differ
    for i in range(prediction_length):
        try:
            worksheet_data.cell(row=(2+len(z_dates)+i), column=3).value = baseline_prediction[i]
        except:
            pass
        
        try:
            worksheet_data.cell(row=(2+len(z_dates)+i), column=6).value = adverse_prediction[i]
        except:
            pass
        
        try:
            worksheet_data.cell(row=(2+len(z_dates)+i), column=7).value = low_prediction[i]
        except:
            pass
        
        # adds prediction dates
        worksheet_data.cell(row=(2+len(z_dates)+i), column=1).value = z_dates[-1].replace(year = z_dates[-1].year + i + 1)
    
    # gets and inserts variable parameters
    variables = model_results.get_variables()
    coefficients = model_results.get_coefficients()
    p_values = model_results.get_p_values()
    
    # writes intercept parameters
    worksheet_summary["A27"] = 'Intercept'
    worksheet_summary["C27"] = p_values[0]
    worksheet_summary["D27"] = coefficients[0]
    
    # writes variable parameters
    for i in range(len(variables)):
        worksheet_summary.cell(row=(i+28), column=1).value = variables[i][0]
        worksheet_summary.cell(row=(i+28), column=2).value = variables[i][1]
        worksheet_summary.cell(row=(i+28), column=3).value = p_values[i + 1]
        worksheet_summary.cell(row=(i+28), column=4).value = coefficients[i + 1]
        
        # add column headers for rolling regression statistics
        worksheet_rr.cell(row=2, column=(i+4)).value = variables[i][0] + " " + variables[i][1] 
        
        # add column headers for time series statistics
        worksheet_time_series.cell(row=2, column=3+i).value = str(variables[i][0] + " " + variables[i][1] )
        
    # writes out times series used in regression
    variable_time_series = model_results.get_variable_time_series()    
    for i in range(len(z_dates)):        
        worksheet_time_series.cell(row=3+i, column=1).value = z_dates[i]
        worksheet_time_series.cell(row=3+i, column=2).value = Z[i]        
        for j in range(len(variables)):
            worksheet_time_series.cell(row=3+i, column=3+j).value = variable_time_series[j][i]
            
    # gets predicted time series for the different scenarios
    # not used in current version of program, but will be needed if functionality for breaking down the effect from each variable on Z forward in time is implemented
    baseline_macro_predictions = model_results.get_baseline_macro_predictions()
    adverse_macro_predictions = model_results.get_adverse_macro_predictions()
    low_macro_predictions = model_results.get_low_macro_predictions()
    
    # macro data predictions for all three scenarios
    #print("Baseline", baseline_macro_predictions)
    #print("Adverse", adverse_macro_predictions)
    #print("Low", low_macro_predictions)
    
    # adds variable stationarity statistics to the output
    i = 0
    variable_stationarity_dict = model_results.get_variable_stationarity()
    for variable in variable_stationarity_dict.keys():
        # joins variable name and variable transformation into one string and writes to excel cell
        worksheet_variables.cell(row=(i+3), column=1).value = " ".join(list(variable))
        
        worksheet_variables.cell(row=(i+3), column=2).value = variable_stationarity_dict[variable]["KPSS"]
        worksheet_variables.cell(row=(i+3), column=3).value = variable_stationarity_dict[variable]["ADF"]        
        i += 1
    
    # gets rolling regression statistics
    RR_statistics = model_results.get_rolling_regression_statistics()
    RR_z = RR_statistics["Z"]
    RR_predicted_z = RR_statistics["Model_yearly_predictions"]
    RR_residuals = RR_statistics["Model_residuals"]
    RR_coefficients_dict = RR_statistics["Coefficients"]
    RR_coefficients_list = list(RR_coefficients_dict.values())
    RR_predicted_upper = RR_statistics["Model_conf_int_upper"]
    RR_predicted_lower = RR_statistics["Model_conf_int_lower"]
    
    # writes rolling regression statistics to excel file
    for i in range(len(RR_z)):  
        # writes data used in visualisation of rolling regression
        worksheet_data.cell(row=(i+2), column=9).value = z_dates[-(len(RR_z) - i)]      
        worksheet_data.cell(row=(i+2), column=10).value = RR_z[i]
        worksheet_data.cell(row=(i+2), column=11).value = RR_predicted_z[i]
        worksheet_data.cell(row=(i+2), column=12).value = RR_predicted_upper[i]
        worksheet_data.cell(row=(i+2), column=13).value = RR_predicted_lower[i]
        
        # writes rolling regression residuals and variable coefficients
        worksheet_rr.cell(row=(i+3), column=1).value = z_dates[-(len(RR_z) - i)]
        worksheet_rr.cell(row=(i+3), column=2).value = RR_residuals[i]
        worksheet_rr.cell(row=(i+3), column=3).value = RR_coefficients_list[i][0]
        
        for j in range(len(variables)):
            worksheet_rr.cell(row=3+i, column=4+j).value = RR_coefficients_list[i][1+j]        
       
    
    # get model statistics
    adjusted_r_square = model_results.get_adjusted_r_squared()
    r_squared = model_results.get_r_squared()
    IFRS9_macro_name = model_results.get_model_name()
    durbin_watson = model_results.get_DW()
    jarque_bera = model_results.get_JB()
    breusch_pagan = model_results.get_BP()
    omni = model_results.get_omni()
    lillefors = model_results.get_lillefors()
    MAE = model_results.get_MAE()
    RMSE = model_results.get_RMSE()
    KPSS = model_results.get_residuals_KPSS()
    ADF = model_results.get_residuals_ADF()
    
    # joins KPSS and ADF results so it can be written to one cell in excel
    residual_stationarity = ", ".join([str(round(KPSS, 4)), str(round(ADF, 4))])
    
    # inserts statistics to output file
    worksheet_summary["B4"] = adjusted_r_square
    worksheet_summary["B6"] = residual_stationarity
    worksheet_summary["B5"] = MAE
    worksheet_summary["D5"] = RMSE
    worksheet_summary["D4"] = IFRS9_macro_name
    worksheet_summary["D6"] = durbin_watson
    worksheet_summary["A10"] = jarque_bera[0]
    worksheet_summary["B10"] = jarque_bera[1]
    worksheet_summary["C10"] = jarque_bera[2]
    worksheet_summary["D10"] = jarque_bera[3]
    worksheet_summary["A14"] = omni[0]
    worksheet_summary["B14"] = omni[1]
    worksheet_summary["C14"] = lillefors[0]
    worksheet_summary["D14"] = lillefors[1]
    worksheet_summary["A18"] = breusch_pagan[0]
    worksheet_summary["B18"] = breusch_pagan[1]
    worksheet_summary["C18"] = breusch_pagan[2]
    worksheet_summary["D18"] = breusch_pagan[3]
        
    # gets rolling regression comparison results
    model_mae = RR_statistics["Model_MAE"]
    model_rmse = RR_statistics["Model_RMSE"]
    as_is_mae = RR_statistics["as_is_MAE"]
    as_is_rmse = RR_statistics["as_is_RMSE"]
    ttc_mae = RR_statistics["ttc_MAE"]
    ttc_rmse = RR_statistics["ttc_RMSE"]
    
    # inserts roling regression statistics to output file
    worksheet_summary["B22"] = model_mae
    worksheet_summary["B23"] = model_rmse
    worksheet_summary["C22"] = as_is_mae
    worksheet_summary["C23"] = as_is_rmse
    worksheet_summary["D22"] = ttc_mae
    worksheet_summary["D23"] = ttc_rmse
    
    # saves and close workbook object
    myworkbook.save(str(OLS_results_ouput_path + OLS_results_output_name))
    myworkbook.close()

def save_model_results(macro_model, variables_info_df, path_dict):
    
    """
            Purpose:
            Saves new model coefficient for a given macro for later deployment to the ECL model

            Inputs:
            macro_model: name of macro model to be saved
            variables_info_df: dataframe with variable names, transformations and coefficients
            path_dict: dictionary with paths to the project model coefficients directory

    """  
    
    # the mapping between macro variable abbreviation and full name
    macro_variable_name_mapping = {'WGDP': 'World GDP', 'ADVGDP': 'Advanced Countries GDP', 'EMEGDP': 'Emerging Countries GDP', 'Oil_price': 'Brent spot oil price Europe', 'USINF': 'US Inflation', 'US 3M': 'US 3 Month Interbank rates', 'SEGDP': 'GDP   Sweden', 'SEINV': 'Investment  Sweden',	'SEEXP': 'Export  Sweden', 'SALMON_PRICE': 'Salmon Price', 'NOGDP': 'Mainland GDP', 'NOINV': 'Mainland Investment',	'NOEXP': 'Traditional Export', 'NOINF':	'Inflation', 'C2_corp': 'Corporates Debt level', 'C2_hh': 'Households debt level', 'HPI': 'House Price', 'no_lr_hh': 'Interest rate to households',	'un': 'Registered unemployment rate', 'i44': 'Exchange rate index I44', 'EURNOK': 'Exchange rate', 'Nibor': 'Nibor 3-month interest rate', 'no_lr_all': 'Interest rate to Corporates', 'rental_price': 'Rental Prices', 'VLCC': 'VLCC spot rate', 'Aframax': 'Aframax spot rate', 'capesize': 'Capesize spot rate', 'Panamax': 'Panamax spot rate', 'fleet_cont': 'Net Container Fleet growth', 'demand_cont': 'World Seaborne Container Trade', 'fleet_gas': 'Total LPG Fleet Development', 'VLGC': 'VLGC spot rate', 'floater_rate': 'Floater spot rate', 'utilization_rate_rig': 'Total Utilization rate rig'}
    
    # loads workbook and sheet
    myworkbook=openpyxl.load_workbook(str(path_dict["Project_coefficients_path"] + 'Macro_models.xlsx'))
    worksheet_model= myworkbook.get_sheet_by_name(macro_model)
    
    # inserts the OLS parameters in the required format
    for i in range(len(variables_info_df.index)):
        
        # special case for the intercept
        if variables_info_df.iloc[i, 0] == "Intercept":
            worksheet_model.cell(row=(i+2), column=1).value = variables_info_df.iloc[i, 0]
            worksheet_model.cell(row=(i+2), column=2).value = "Intercept"
            
        # special case for dummies
        elif "Dummy" in variables_info_df.iloc[i, 0]:
            dummy_list = variables_info_df.iloc[i, 0].split(" ")
            worksheet_model.cell(row=(i+2), column=1).value = dummy_list[0]
            worksheet_model.cell(row=(i+2), column=2).value = dummy_list[0]
            worksheet_model.cell(row=(i+2), column=3).value = dummy_list[1]            
        else:
            worksheet_model.cell(row=(i+2), column=1).value = variables_info_df.iloc[i, 0]
            worksheet_model.cell(row=(i+2), column=2).value = macro_variable_name_mapping[variables_info_df.iloc[i, 0]]
            worksheet_model.cell(row=(i+2), column=3).value = variables_info_df.iloc[i, 1]
            
        worksheet_model.cell(row=(i+2), column=4).value = variables_info_df.iloc[i, 3]
    
    myworkbook.save(str(path_dict["Project_coefficients_path"] + 'Macro_models.xlsx'))
    myworkbook.close()
    
    # for updating the sheet with what models have been changed
    changed_models_df = pd.read_excel(str(path_dict["Project_coefficients_path"] + 'Changed_models.xlsx'), header=0)
    timestamp = time.strftime("%d-%m-%Y")
    changed_models_df.loc[changed_models_df.Model == macro_model, 'Changed'] = "Yes"
    changed_models_df.loc[changed_models_df.Model == macro_model, 'Date'] = timestamp
    changed_models_df.to_excel(str(path_dict["Project_coefficients_path"] + 'Changed_models.xlsx'))
    
def deploy_new_models(path_dict, project_name):
    """
            Purpose:
            Saves new model coefficients to the ECL model

            Inputs:
            path_dict: dictionary with paths to macro model parameters file in ECL project and path to the saved macro model parameters in the macro tool project  
            project_name: name of macro model tool project

    """    
    #gets directory where model coefficients for ECL model are stored
    ECL_macro_model_dir = str("/".join(path_dict["Macro_models"].split("/")[:-1]) + "/")
    output_name = str(ECL_macro_model_dir + project_name + ".xlsx")
    
    # copies project model parameters to ECL model
    shutil.copy(str(path_dict["Project_coefficients_path"] + "Macro_models.xlsx"), output_name)
    
    print("Model parameters successfully deployed to ECL model")
    print("Remember to update file paths in the ECL model before using the new macro models")
                               
def trim_time_series(z_time_series, variable_time_series_dict, dummy_dict, z_dates, adverse_variable_time_series_dict, low_variable_time_series_dict):
    """
            Purpose:
            Shorten the time series in order for them to have the same length.
            Specifically it shortens z_time_series and all historical time series in variable_time_series_dict to same length
            Also shortens all predicted future time series in variable_time_series_dict, adverse_variable_time_series_dict and low_variable_time_series_dict individually to same length

            Inputs:
            z_time_series: list of the z values
            variable_time_series_dict: dictionary with key=variable names, value=list of variable values
            dummy_dict: dictionary with key=dummy names, value=list of dummy values
            z_dates: list of dates to be used in the regressions
            adverse_variable_time_series_dict: dictionary with key=variable names, value=list of variable values
            low_variable_time_series_dict: dictionary with key=variable names, value=list of variable values

            Outputs:
            A tuple with trimmed actuals time series, full projections series, dates used for actuals and variable names               

    """
    
    # to determine what time series is the shortest
    shortest_time_series = len(z_time_series)
    
    variable_time_series_list = []
    prediction_time_series_list = []
    variable_names = []
    
    # loops through all variables in baseline
    for variable in variable_time_series_dict:
        variable_names.append(variable)
        
        # extracts the variable time series
        variable_dict = variable_time_series_dict[variable]
        time_series_tuples = variable_dict.values() 
        
        variable_time_series = list(map(lambda x: x[0] if x[1] == "_Actuals" else None, time_series_tuples))
        variable_time_series = list(filter(None, variable_time_series))
        variable_time_series_list.append(variable_time_series)
        
        prediction_time_series = list(map(lambda x: x[0] if x[1] == "_Projections" else None, time_series_tuples))
        prediction_time_series = list(filter(None, prediction_time_series))
        prediction_time_series_list.append(prediction_time_series)
        
        # sets the variable to the shortest time series
        if len(variable_time_series) < shortest_time_series:
            shortest_time_series = len(variable_time_series)
            
    # if dummy_dict contain dummy time series, these are appended to return list
    if bool(dummy_dict):
        for dummy in dummy_dict:
            variable_time_series_list.append(dummy_dict[dummy])
            variable_names.append(dummy)
    
    # trims time series in order to have the same length
    z_time_series = z_time_series[len(z_time_series) - shortest_time_series:]
    variable_time_series_list = list(map(lambda x: x[len(x)-shortest_time_series:], variable_time_series_list))
    z_dates = z_dates[len(z_dates) - shortest_time_series:]
    
    # do the same operation to get adverse projections
    adverse_prediction_time_series_list = []
    for variable in adverse_variable_time_series_dict:
        
        # extracts the variable time series
        variable_dict = adverse_variable_time_series_dict[variable]
        time_series_tuples = variable_dict.values()
        
        adverse_prediction_time_series = list(map(lambda x: x[0] if x[1] == "_Projections" else None, time_series_tuples))
        adverse_prediction_time_series = list(filter(None, adverse_prediction_time_series))
        adverse_prediction_time_series_list.append(adverse_prediction_time_series)
    
    # do the same operation to get low projections
    low_prediction_time_series_list = []
    for variable in low_variable_time_series_dict:
        
        # extracts the variable time series
        variable_dict = low_variable_time_series_dict[variable]
        time_series_tuples = variable_dict.values()
        
        low_prediction_time_series = list(map(lambda x: x[0] if x[1] == "_Projections" else None, time_series_tuples))
        low_prediction_time_series = list(filter(None, low_prediction_time_series))
        low_prediction_time_series_list.append(low_prediction_time_series)

    
    # trims prediction time series in order to have the same length
    shortest_prediction = len(prediction_time_series_list[0])
    for variable in prediction_time_series_list:
        if len(variable) < shortest_prediction:
            shortest_prediction = len(variable)
            
    # cuts of the 'tail' of the predictions longer than the shortest one
    prediction_time_series_list = list(map(lambda x: x[:shortest_prediction], prediction_time_series_list))
            
    # trims adverse predictions series in order to have the same length
    shortest_adverse_prediction = len(adverse_prediction_time_series_list[0])
    for variable in adverse_prediction_time_series_list:
        if len(variable) < shortest_adverse_prediction:
            shortest_adverse_prediction = len(variable)
    
    # cuts of the 'tail' of the adverse predictions longer than the shortest one
    adverse_prediction_time_series_list = list(map(lambda x: x[:shortest_adverse_prediction], adverse_prediction_time_series_list))
    
    # trims low predictions series in order to have the same length
    shortest_low_prediction = len(low_prediction_time_series_list[0])
    for variable in low_prediction_time_series_list:
        if len(variable) < shortest_low_prediction:
            shortest_low_prediction = len(variable)
    
    # cuts of the 'tail' of the low predictions longer than the shortest one
    low_prediction_time_series_list = list(map(lambda x: x[:shortest_low_prediction], low_prediction_time_series_list))
                                       
    # create dummy time series with value 0 and length equal to the variable time series. Do for both baseline, adverse and low
    # this in order to have dummy time series with the same length as the other time series, but without any effect
    if bool(dummy_dict):
        for dummy in dummy_dict:
            prediction_time_series_list.append([0] * shortest_prediction)
            adverse_prediction_time_series_list.append([0] * shortest_adverse_prediction)
            low_prediction_time_series_list.append([0] * shortest_low_prediction)
    

    return z_time_series, variable_time_series_list, variable_names, z_dates, prediction_time_series_list, adverse_prediction_time_series_list, low_prediction_time_series_list
    
            
def do_extended_OLS(z_time_series, variable_time_series_list, variable_names=None, baseline_predictions=None, adverse_predictions=None, low_predictions=None):
    """
               Purpose:
               Return extended OLS regression statistics
               Note that if variable_names is given, baseline_predictions, adverse_predictions and low_predictions must also be given

               Inputs:
               z_time_series: list with Z time series to be modelled
               variable_time_series_dict: dict with key=variable name, value=dict with key=timestamp, value = tuple(time series value, actuals/projected(string))
               variable_names = list with variables names
               Outputs:
               
                
        """

    # converts explanatory variables to an array and adds a constant
    explanatory_variables_array = np.array(variable_time_series_list).T
    explanatory_variables_array = sm.add_constant(explanatory_variables_array, has_constant='add')
    
    # generate the base model
    model = sm.OLS(z_time_series, explanatory_variables_array).fit()
    
    # correct the parameters for heteroscedasticity and autocorrelation
    model = model.get_robustcov_results(cov_type='HAC', maxlags=1)
    
    # creates an object of type OLS_result_object to store OLS results
    OLS_results = OLS_result_object()
    OLS_results.set_model(model)
    OLS_results.set_coefficients((model.params).tolist()) 
    OLS_results.set_z(z_time_series)
    OLS_results.set_variable_time_series(variable_time_series_list)
    
    # if the variables under are given, more extensive statistical data is calculated and saved to the OLS results object
    if variable_names != None and baseline_predictions != None and adverse_predictions != None and low_predictions != None:
        
        # calculates stationarity statistics for the variable time series
        stationarity_dict = {}
        for i in range(len(variable_time_series_list)):
            if "Dummy" not in variable_names[i]:
                stationarity_dict[variable_names[i]] = {"KPSS": sm.tsa.stattools.kpss(variable_time_series_list[i], regression='c', nlags='auto', store=True)[1], 
                                                        "ADF": sm.tsa.stattools.adfuller(variable_time_series_list[i], regression='c', store=True)[1]}
    
        OLS_results.set_variable_stationarity(stationarity_dict)
        
        # saves OLS result parameters to result object
        OLS_results.set_adjusted_r_squared(model.rsquared_adj)    
        OLS_results.set_r_squared(model.rsquared)
        OLS_results.set_p_values(list(model.pvalues))
        OLS_results.set_residuals(model.resid)
        predictions = model.predict(explanatory_variables_array)
        OLS_results.set_predictions(predictions)
        
        # set baseline predictions based on forward in time macro variables
        OLS_results.set_baseline_macro_predictions(baseline_predictions)
        baseline_explanatory_variables_array = np.array(baseline_predictions).T
        baseline_explanatory_variables_array = sm.add_constant(baseline_explanatory_variables_array, has_constant='add')
        OLS_results.set_baseline_predictions(model.predict(baseline_explanatory_variables_array))
        
        # set adverse predictions based on adverse forward in time macro variables
        OLS_results.set_adverse_macro_predictions(adverse_predictions)
        adverse_explanatory_variables_array = np.array(adverse_predictions).T
        adverse_explanatory_variables_array = sm.add_constant(adverse_explanatory_variables_array, has_constant='add')
        OLS_results.set_adverse_predictions(model.predict(adverse_explanatory_variables_array))
        
        # set low predictions based on low forward in time macro variables
        OLS_results.set_low_macro_predictions(low_predictions)
        low_explanatory_variables_array = np.array(low_predictions).T
        low_explanatory_variables_array = sm.add_constant(low_explanatory_variables_array, has_constant='add')
        OLS_results.set_low_predictions(model.predict(low_explanatory_variables_array))
        
        # calculates and sets RMSE and MAE
        MSE_error_list = []
        MAE_error_list = []
        for i in range(len(z_time_series)):
            MSE_error_list.append((z_time_series[i] - OLS_results.get_predictions()[i])**2)
            MAE_error_list.append(abs(z_time_series[i] - OLS_results.get_predictions()[i]))
        
        RMSE = (statistics.mean(MSE_error_list))**0.5
        MAE = statistics.mean(MAE_error_list)
        OLS_results.set_RMSE(RMSE)
        OLS_results.set_MAE(MAE)
        
        # calculates and set confidence intervals
        prob__int_addition = norm.ppf(0.0975)
        model_conf_int_upper = []
        model_conf_int_lower = []
        for i in range(len(predictions)):
            model_conf_int_upper.append(predictions[i] + RMSE * prob__int_addition)
            model_conf_int_lower.append(predictions[i] - RMSE * prob__int_addition)
        
        OLS_results.set_confidence_intervals(lower=model_conf_int_lower, upper=model_conf_int_upper)
        
        # runs statistical tests and saves result to result object
        OLS_results.set_DW(sm.stats.stattools.durbin_watson(OLS_results.get_residuals()))
        OLS_results.set_JB(sm.stats.stattools.jarque_bera(OLS_results.get_residuals()))
        OLS_results.set_omni(sm.stats.stattools.omni_normtest(OLS_results.get_residuals()))
        OLS_results.set_BP(sm.stats.diagnostic.het_breuschpagan(OLS_results.get_residuals(), sm.add_constant(explanatory_variables_array)))
        OLS_results.set_lillefors(sm.stats.diagnostic.kstest_normal(OLS_results.get_residuals()))
           
    return OLS_results
          

def rolling_regression(z_time_series, variable_time_series_list, years_to_roll):
    """
                Purpose:
                Do an out of sample rolling regression to see how the model will behave out of sample and over time

                Inputs:
                z_time_series: list of the Z values
                variable_time_series_list: list where the elements are lists of variable values
                years_to_roll: integer determining how many years to do rolling regression for             
            
                Output:
                dictionary with rolling regression results
        """
    
    # creates the required lists and dictionaries
    model_residuals_list = []
    model_predictions_list = []
    model_predictions_full_time_series = []    
    model_coefficients_dict = {}
    model_conf_int_upper = []
    model_conf_int_lower = []    
    as_is_residuals_list = []
    as_is_prediction_list = []    
    ttc_residuals_list = []         
    z_list = []
    

    # doing rolling regression equal to variable years_to_roll
    for i in range (years_to_roll, 0, -1):        
        
        # calculate residual for as-is
        as_is_residuals_list.append(z_time_series[-(i + 1)] - z_time_series[-(i)])
        as_is_prediction_list.append(z_time_series[-(i + 1)])
        
        # sets TTC residuals
        ttc_residuals_list.append(z_time_series[-(i)])
              
        # estimates new OLS model based on shorten time series
        reduced_model = do_extended_OLS(z_time_series[:-i], list(map(lambda x: x[: -i], variable_time_series_list)))
        
        # special case for the last round of rolling regression since list[-0] is not possible
        if i == 1:
            prediction_variables = variable_time_series_list
        else:
            prediction_variables = list(map(lambda x: x[: -i + 1], variable_time_series_list))
                        
        #makes prediction based on on year ahead in time macro data
        prediction = reduced_model.calculate_predictions(prediction_variables) 
        
        # gets prediction for the last year and compares to actual Z
        model_predictions_list.append(prediction[-1])
        model_residuals_list.append(prediction[-1] - z_time_series[-i])
        
        # sets predictions for the entire prediction period
        model_predictions_full_time_series.append(prediction)
        
        # appends the coefficients of the variables for each year of rolling regression
        model_coefficients_dict[i] = reduced_model.get_coefficients()       
        
        # appends z value used in rolling regression
        z_list.append(z_time_series[-i])
    
    #calculate MAE and RMSE for the three models
    as_is_MAE = statistics.mean(list(map(lambda x: abs(x), as_is_residuals_list)))
    as_is_RMSE = (statistics.mean(list(map(lambda x: x**2, as_is_residuals_list))))**(0.5)
    
    ttc_MAE = statistics.mean(list(map(lambda x: abs(x), ttc_residuals_list)))
    ttc_RMSE = (statistics.mean(list(map(lambda x: x**2, ttc_residuals_list))))**(0.5)
    
    model_MAE = statistics.mean(list(map(lambda x: abs(x), model_residuals_list)))
    model_RMSE = (statistics.mean(list(map(lambda x: x**2, model_residuals_list))))**(0.5)
    
    # estimates the confidence intervalls
    prob__int_addition = norm.ppf(0.0975)
    model_conf_int_upper = []
    model_conf_int_lower = []
    for i in range(len(model_predictions_list)):
        model_conf_int_upper.append(model_predictions_list[i] + model_RMSE * prob__int_addition)
        model_conf_int_lower.append(model_predictions_list[i] - model_RMSE * prob__int_addition)
        
    return_dict = {"Coefficients": model_coefficients_dict, 
                   "Model_MAE": model_MAE, "Model_RMSE": model_RMSE, 
                   "as_is_MAE": as_is_MAE, "as_is_RMSE": as_is_RMSE, 
                   "ttc_MAE": ttc_MAE, "ttc_RMSE": ttc_RMSE,
                   "Model_residuals": model_residuals_list, 
                   "Model_yearly_predictions": model_predictions_list,
                   "Model_full_predictions": model_predictions_full_time_series,
                   "Model_conf_int_upper": model_conf_int_upper,
                   "Model_conf_int_lower": model_conf_int_lower,
                   "Z": z_list
                  }
    
    return return_dict
    
class OLS_result_object:
    """
    A class mainly for storing the regression results.
    Is is also utilized in the rolling regression to make that code neater
    """
    
    def __init__(self):
        self.dates = None
        self.Z = None
        self.IFRS9_macro_model_name = None
        self.variables = None
        self.projected_z_KPSS = None
        self.residuals_KPSS = None
        self.residuals_ADF = None
        self.model = None
        self.coefficients = {}
        self.p_values = {}
        self.residuals = []
        self.variable_stationarity = {}
        self.adjusted_r_squared = None
        self.r_squared = None
        self.residuals_mean = 0
        self.DW_statistics = None
        self.JB_Statistics = None
        self.omni_statistics = None
        self.BP_statistics = None
        self.lillefors_statistics = None
        self.rolling_regression_statistics = None
        self.confidence_intervals = None
        self.predictions = None
        self.MAE = None
        self.RMSE = None
        self.baseline_predictions = None
        self.adverse_predictions = None
        self.low_predictions = None
        self.variable_time_series = None
        self.baseline_macro_predictions = None
        self.adverse_macro_predictions = None
        self.low_macro_predictions = None
        
    def set_adverse_macro_predictions(self, adverse_macro_predictions):
        self.adverse_macro_predictions = adverse_macro_predictions
        
    def get_adverse_macro_predictions(self):
        return self.adverse_macro_predictions
    
    def set_low_macro_predictions(self, low_macro_predictions):
        self.low_macro_predictions = low_macro_predictions
        
    def get_low_macro_predictions(self):
        return self.low_macro_predictions
    
    def set_baseline_macro_predictions(self, baseline_macro_predictions):
        self.baseline_macro_predictions = baseline_macro_predictions
        
    def get_baseline_macro_predictions(self):
        return self.baseline_macro_predictions
        
    def set_variable_time_series(self, variable_time_series):
        self.variable_time_series = variable_time_series
        
    def get_variable_time_series(self):
        return self.variable_time_series
        
    def set_baseline_predictions(self, predictions):
        self.baseline_predictions = predictions
        
    def get_baseline_predictions(self):
        return self.baseline_predictions
    
    def set_adverse_predictions(self, predictions):
        self.adverse_predictions = predictions
        
    def get_adverse_predictions(self):
        return self.adverse_predictions
    
    def set_low_predictions(self, predictions):
        self.low_predictions = predictions
        
    def get_low_predictions(self):
        return self.low_predictions
        
    def set_MAE(self, MAE):
        self.MAE = MAE
        
    def get_MAE(self):
        return self.MAE
    
    def set_RMSE(self, RMSE):
        self.RMSE = RMSE
    
    def get_RMSE(self):
        return self.RMSE
        
    def set_z_KPSS(self, KPSS):
        self.projected_z_KPSS = KPSS
        
    def get_z_KPSS(self):
        return self.projected_z_KPSS
    
    def set_residual_KPSS(self, KPSS):
        self.residuals_KPSS = KPSS
        
    def get_residual_KPSS(self):
        return self.residuals_KPSS
    
    def set_variables(self, names):
        self.variables = names
        
    def get_variables(self):
        return self.variables
    
    def set_model_name(self, name):
        self.IFRS9_macro_model_name = name
        
    def get_model_name(self):
        return self.IFRS9_macro_model_name
    
    def set_z(self, Z):
        self.z = Z
                       
    def get_z(self):
        return self.z                       
                       
    def set_dates(self, dates):
        self.dates = dates
        
    def get_dates(self):
        return self.dates
    
    def set_model(self, model):
        self.model = model
        
    def set_coefficients(self, coefficients_dict):
        self.coefficients = coefficients_dict
        
    def get_coefficients(self):
        return self.coefficients
    
    def set_p_values(self, p_values_dict):
        self.p_values = p_values_dict
        
    def get_p_values(self):
        return self.p_values
    
    # function to both set residuals and stationarity statistics for the model residuals
    def set_residuals(self, residuals):
        self.residuals_KPSS= sm.tsa.stattools.kpss(residuals, regression='c', nlags='auto', store=True)[1]
        self.residuals_ADF= sm.tsa.stattools.adfuller(residuals, regression='c', store=True)[1]
        self.residuals_mean = statistics.mean(residuals)
        self.residuals = residuals
        
    def get_residuals(self):
        return self.residuals
    
    def get_residuals_mean(self):
        return self.residuals_mean
        
    def get_residuals_KPSS(self):
        return self.residuals_KPSS
    
    def get_residuals_ADF(self):
        return self.residuals_ADF 
    
    def set_variable_stationarity(self, variable_stationarity_dict):
        self.variable_stationarity = variable_stationarity_dict
        
    def get_variable_stationarity(self):
        return self.variable_stationarity
    
    def set_adjusted_r_squared(self, adjusted_r_squared):
        self.adjusted_r_squared = adjusted_r_squared
        
    def get_adjusted_r_squared(self):
        return self.adjusted_r_squared
    
    def set_r_squared(self, r_squared):
        self.r_squared = r_squared
        
    def get_r_squared(self):
        return self.r_squared
    
    def set_predictions(self, predictions):
        self.predictions = predictions
        
    def get_predictions(self):
        return self.predictions
    
    def calculate_predictions(self, variable_time_series_list):
        explanatory_variables_array = np.array(variable_time_series_list).T
        explanatory_variables_array = sm.add_constant(explanatory_variables_array)
        
        predictions = self.model.get_prediction(explanatory_variables_array)
        mean = predictions.predicted_mean           
        
        return mean
    
    def set_DW(self, DW_statistics):
        self.DW_statistics = DW_statistics
        
    def get_DW(self):
        return self.DW_statistics
    
    def set_JB(self, JB_statistics):
        self.JB_statistics = JB_statistics
        
    def get_JB(self):
        return self.JB_statistics
        
    def set_omni(self, omni_statistics):
        self.omni_statistics = omni_statistics
        
    def get_omni(self):
        return self.omni_statistics
    
    def set_BP(self, BP_statistics):
        self.BP_statistics = BP_statistics
        
    def get_BP(self):
        return self.BP_statistics
    
    def set_lillefors(self, lillefors_statistics):
        self.lillefors_statistics = lillefors_statistics
        
    def get_lillefors(self):
        return self.lillefors_statistics
    
    def set_rolling_regression_statistics(self, comparison_dict):
        self.rolling_regression_statistics = comparison_dict
    
    def get_rolling_regression_statistics(self):
        return self.rolling_regression_statistics
    
    def set_confidence_intervals(self, lower, upper):
        self.confidence_intervals = {"Lower": lower, "Upper": upper}
        
    def get_confidence_intervals(self):
        return self.confidence_intervals
    



def generate_OLS_models(path_dict, macro_model, variables, max_variables=3, max_dummies=1, max_one_variable=1, significance_level=0.05, r_cutoff_level=0.65, max_lag=3, Y1_predict_aim=None):
    """
        Purpose: 
            Generate suggested OLS models for the specified macro model.
            Outputs both a list of suggested models without dummies, and adds dummies to those suggestions to output a list of those models with dummies as well
            
        Inputs:
            path_dict: dictionary with paths to the required files
            macro_model: string specifyng what IFRS macro model to estimate
            variables: dataframes with variables to be tried for simulation
            max_variables: integer specifying maximum number of variables to try
            max_dumies: integer specifying maximum number of dummies to try
            max_one_variables: integer specifying maximum number of transformations of the same variable to try
            significance_level: float specifying the significance level when doing OLS regressions
            r_cutoff_level: float specifying the cutoff level in regards to r squared of the OLS models
            Y1_predict_aim: percentile aim for first year of predictions
            
        Outputs:
            Macro_models_without_dummies.xlsx: suggested models without dummies
            Macro_models_with_dummies.xlsx: suggested models with dummies
        
    """
    
    # sets a starting time to give timing information to user
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    start_time = time.time()
    
    # gets Z time series to be modelled
    z_credit_cycle = get_Z_values(path_dict, macro_model)    
    z_date_list = z_credit_cycle["z_date_list"]
    z_values = z_credit_cycle["z_values"]
    
    # if goal for Y1 prediction is set, this value is added to Z_values and a new date is added to z_date_list
    if Y1_predict_aim:
        z_values.append(norm.ppf(Y1_predict_aim))
        new_date = z_date_list[-1] + pd.tseries.offsets.DateOffset(years=1)
        z_date_list.append(new_date)
    
    statistics_dict = {}
    # stores simulation parameters
    statistics_dict["Aim Y1"] = Y1_predict_aim    
    statistics_dict["Z"] = z_values
    statistics_dict["Dates"] = z_date_list
    statistics_dict["Max variables"] = max_variables
    statistics_dict["Max dummies"] = max_dummies
    statistics_dict["Max lag"] = max_lag
    statistics_dict["Max one variable"] = max_one_variable
    statistics_dict["Significance level"] = significance_level
    statistics_dict["R level"] = r_cutoff_level
    
    # gets variable names from variable info dataframe
    statistics_dict["Variables"] = list(variables.iloc[:,0])
        
    # imports, validates and cleans macro data
    macro_data_dict = import_and_clean_macro_input(path_dict, "Baseline_sheet")
    if macro_data_dict:
        dnb_markets_macro_transformed = macro_data_dict["dnb_markets_macro_transformed"]
        us_inflation_actuals = macro_data_dict["us_inflation_actuals"]
        us_inflation_projections = macro_data_dict["us_inflation_projections"]
        nor_inflation_actuals = macro_data_dict["nor_inflation_actuals"]
        nor_inflation_projections = macro_data_dict["nor_inflation_projections"]
        dates = macro_data_dict["dates"]
        index_years = macro_data_dict["index_years"]
        
        variable_dict = {}
        variable_sign_dict = {}
        macro_model_variable_list = variables
        variable_transform_statistics = {}
        
        #iterates through all the variables to be used in simulation
        for index, row in macro_model_variable_list.iterrows():

            # sets the correct variable for doing Ln
            if row[2].upper() == "Y":
                do_ln = True
            else:
                do_ln = False

            # set on the correct inflation data for transformation
            inflation_type = row[1]
            try:
            
                if inflation_type == None:
                    inflation_data_actuals = None
                    inflation_data_projections = None
                    inflation_type = None
                elif inflation_type.upper() == "US":
                    inflation_data_actuals = us_inflation_actuals
                    inflation_data_projections = us_inflation_projections
                elif inflation_type.upper() == "NOR":
                    inflation_data_actuals = nor_inflation_actuals
                    inflation_data_projections = nor_inflation_projections
                else:
                    inflation_data_actuals = None
                    inflation_data_projections = None
                    inflation_type = None
            except:
                inflation_data_actuals = None
                inflation_data_projections = None
                inflation_type = None
                
            # sets the expected effect from variable on credit cycle (positiv or negativ)
            variable_sign_dict[row[0]] = row[3]
                
            # extracts actuals and projected macro data and transforms to float
            macro_data_actuals = dnb_markets_macro_transformed[row[0]+"_Actuals"]
            macro_data_actuals = pd.Series(list(map(lambda x: float(x), macro_data_actuals)))
            macro_data_projections = dnb_markets_macro_transformed[row[0] + "_Projections"]
            macro_data_projections = pd.Series(list(map(lambda x: float(x), macro_data_projections)))
            
            # formats the actuals and projections to list of tuples. Required format for rest of program
            macro_data_list_actuals = (dates, macro_data_actuals)
            macro_data_list_projections = (dates, macro_data_projections)
 
            # transforms the variable to all combinations
            variable_dict[row[0]] = do_variable_transformations(variable_name=row[0],macro_variable_actuals=macro_data_list_actuals,
                                                                macro_variable_projections=macro_data_list_projections,
                                                                do_ln=do_ln, inflation_data_actuals=inflation_data_actuals,
                                                                inflation_data_projections=inflation_data_projections,
                                                                inflation_type=inflation_type, inflation_metadata=index_years, max_lag=max_lag)
            variable_transform_statistics[row[0]] = len(variable_dict[row[0]])
            
        # append number of transformations for each variable to simulation statistics
        statistics_dict["Variable transformations"] = variable_transform_statistics
        
        # trimes the transformed variables to be the same dates as the Z to be modelled
        trimmed_variable_dict = {}
        for variable in variable_dict:
            trimmed_variable_dict[variable] = {}
            for transformation in variable_dict[variable]:
                trimmed_variable_dict[variable][transformation] = {}
                for variable_date in variable_dict[variable][transformation].keys():
                    if variable_date in z_date_list:
                        trimmed_variable_dict[variable][transformation][variable_date] = variable_dict[variable][transformation][variable_date]
        
        # makes one large dictionary with transformed values and key variable+transformation.
        # will be used for looking up values for OLS regression
        flattened_variable_dict = {}
        for variable in trimmed_variable_dict:
            for transformation in trimmed_variable_dict[variable]:
                    flattened_variable_dict[variable + " " + transformation] = trimmed_variable_dict[variable][transformation]


        # combines the variables with transformations of the same variable
        combined_variable_transformation_dict = {}   
        for variable in trimmed_variable_dict:
            combined_variable_transformation_dict[variable] = combine(trimmed_variable_dict[variable], max_one_variable, variable)        
        
        # combines the different variables in all possible ways
        combined_variables_list = combine(combined_variable_transformation_dict, max_variables)
       
        # combines the different transformations of variables in all possible ways
        combined_transformation_list = []
        for variable_combination in combined_variables_list:
            
            # checks if there is any combination to be done
            if len(variable_combination) > 1:
                transformations_list = []
                for variable in variable_combination:
                    transformations_list.append(combined_variable_transformation_dict[variable[0]])                
                
                combined_transformation_list.extend(combine_lists(transformations_list))
            else:
                combined_transformation_list.extend(combined_variable_transformation_dict[variable_combination[0]])
            
        #removes duplicates
        combined_transformation_list = list(dict.fromkeys(combined_transformation_list))
        
        # adds number of variable combinations for OLS simualtions
        statistics_dict["Combinations"] = len(combined_transformation_list)
       
        # generate dummies and combine them in all possible ways
        dummy_dict = generate_dummies(z_date_list, max_dummy_length=3)
        combined_dummies_list = combine(dummy_dict, max_dummies)
        
        # adds number of dummy combinations for OLS simualtions
        statistics_dict["Dummy combinations"] = len(combined_dummies_list)
        
        # an object to fetch the time series of the variable transformations
        time_series_object = TimeSeriesLookup(variables_dict=flattened_variable_dict, dummy_dict=dummy_dict, Z=z_values)
                
        # adds additional variables needed for OLS regresssion
        for i in range(len(combined_transformation_list)):
            combined_transformation_list[i] = (combined_transformation_list[i], significance_level, time_series_object, r_cutoff_level, True, variable_sign_dict)
        
        OLS1_time = time.time()
        print("Doing 1st round of regressions...")
        # apply OLS regression to all combinations
        pool = mp.Pool()        
        max_mp_size = 50000
        # to speed up the program, the combinations are broken into chunks of 50 000. 
        # this creates less overhead for the CPU when multiprocessing.
        if len(combined_transformation_list) <= max_mp_size:
            results_no_dummies = list(pool.map(do_OLS, combined_transformation_list))
        else:
            results_no_dummies = []
            # find number of chunks, and the remainder not divisable by 50'000
            remainder = int(len(combined_transformation_list)%max_mp_size)
            number_of_chunks = int((len(combined_transformation_list) - remainder)/max_mp_size)
            
            # do OLS for the chunks of combinations
            last_index = 0
            for i in range(1, 1+number_of_chunks):
                result_list = list(pool.map(do_OLS, combined_transformation_list[last_index: int(i*max_mp_size)]))
                results_no_dummies.extend(result_list)
                last_index = i * max_mp_size
            
            # do OLS for the remainder of the combinations
            if remainder != 0:
                results_no_dummies.extend(list(pool.map(do_OLS, combined_transformation_list[len(combined_transformation_list) - (remainder+1): -1])))

        print("Done with 1st round of regressions in", time.time() - OLS1_time)
        
        pool.close()
        pool.join()
        
        # filters out None returned for failed regressions
        results_no_dummies = list(filter(None, results_no_dummies))
        
        # saves number of non-failed regression from first round of simulations
        statistics_dict["Number of no dummies"] = len(results_no_dummies)
        
        #sorts list based on adjusted r-square
        results_no_dummies.sort(key= lambda x: x[2], reverse=True)
        
        # shortens the list if number of results are too many
        max_no_dummies = 1000
        if len(results_no_dummies) > max_no_dummies:
            results_no_dummies = results_no_dummies[:max_no_dummies]
        
        # makes a list of the combinations returned from the regression with list comprehension
        combinations = [tuple(result[0]) for result in results_no_dummies]
       
        # converts dummy combinations into correct format
        combined_dummies_list = list(map(lambda x: sum(x, ()) if len(x) > 1 else x, combined_dummies_list))
        
        # joins promising OLS models with dummies
        combined_variables_and_dummies_list = combine_lists([combinations, combined_dummies_list])
        
        # adds number of simualtions for round two to simulation statistics
        statistics_dict["With dummy combinations"] = len(combined_variables_and_dummies_list)
        
        # adds required data to the tuple to be sent to OLS regression function
        for i in range(len(combined_variables_and_dummies_list)):
            combined_variables_and_dummies_list[i] = (combined_variables_and_dummies_list[i], significance_level, time_series_object, r_cutoff_level, False, variable_sign_dict)
        
        # apply OLS regression to combination of variables and dummies
        OLS2_time = time.time()
        print("Doing 2nd round of regressions...")
        pool = mp.Pool() 
        # does the same chunking as previously in order to speed up execution
        if len(combined_variables_and_dummies_list) <= max_mp_size:
            results_with_dummies = list(pool.map(do_OLS, combined_variables_and_dummies_list))
        else:
            results_with_dummies = []
            remainder = int(len(combined_variables_and_dummies_list)%max_mp_size)
            number_of_chunks = int((len(combined_variables_and_dummies_list) - remainder)/max_mp_size)
            
            last_index = 0
            for i in range(1, 1+number_of_chunks):
                result_list = list(pool.map(do_OLS, combined_variables_and_dummies_list[last_index: int(i*max_mp_size)]))
                results_with_dummies.extend(result_list)
                last_index = i * max_mp_size
            
            if remainder != 0:
                results_with_dummies.extend(list(pool.map(do_OLS, combined_transformation_list[len(combined_transformation_list) - (remainder+1): -1])))
        
        pool.close()
        pool.join()
        
        print("Done with 2nd round of regressions in", time.time() - OLS2_time)
        
        # filters out None returned for failed regressions
        results_with_dummies = list(filter(None, results_with_dummies))
        
        # adds number of non-failed regressions from round two to simulation statistics
        statistics_dict["Number with dummies"] = len(results_with_dummies)

        #sorts list based on adjusted r-square
        results_with_dummies.sort(key= lambda x: x[2], reverse=True)
        
        # shortens the list if number of results are too many
        max_with_dummies = 10000
        if len(results_with_dummies) > max_with_dummies:
            results_with_dummies = results_with_dummies[:max_with_dummies]

        used_time = time.time() - start_time
        statistics_dict["Time to do simulations"] = used_time
        
        # add number of total regressions done to simulation statistics
        statistics_dict["Number of simulations"] = len(combined_transformation_list) + len(combined_variables_and_dummies_list)

        # makes the output more readable
        results_no_dummies = list(map(lambda x: (str(x[0]).replace("(", "").replace(")", "").replace("'", ""), str(x[1]).replace("'", "")[1:-1], x[2], x[3], x[4]), results_no_dummies))
        
        # makes the output more readable
        results_with_dummies = list(map(lambda x: (str(x[0]).replace("(", "").replace(")", "").replace("'", ""), str(x[1]).replace("'", "")[1:-1], x[2], x[3], x[4]), results_with_dummies))  
                                   
        # calls function to write OLS suggestion results to excel
        write_OLS_suggestions_to_excel(results_no_dummies=results_no_dummies, results_with_dummies=results_with_dummies,
                                       path_dict=path_dict, macro_model=macro_model, statistics_dict=statistics_dict)
        
def write_OLS_suggestions_to_excel(results_no_dummies, results_with_dummies, path_dict, macro_model, statistics_dict):
    """
    Purpose:
    Write results from OLS suggestion simulation to a copy of the OLS suggestions template saved in the project parameters folder on the project. 
    
    Inputs:
    results_no_dummies: list with successful OLS results without dummies. Elements are dummies with the output from the function do_OLS()
    results_with_dummies: list with successful OLS results with dummies. Elements are dummies with the output from the function do_OLS()
    path_dict: dictionary with paths to template and output directory
    macro_model: name of macro model simulation has been done for
    statistics_dict: dictionary with simulation statistics
    """
    
    # gets and uses timestamp to create a unique name for output file
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    template_path = path_dict["OLS_suggestions_template"]
    output_name = str(macro_model + "/" + timestamp + ".xlsx")
    
    # gets output path
    OLS_suggestions_ouput_path = path_dict["OLS_simulation_output_path"]
    
    # checks if this macro model has directory. If not, a directory is created
    model_output_directory = []
    for file in os.listdir(OLS_suggestions_ouput_path):
        model_output_directory.append(file)
        
    if macro_model not in model_output_directory:
        os.mkdir(str(OLS_suggestions_ouput_path + macro_model))
    
    # copies template and change name and directory location
    shutil.copy(template_path, str(OLS_suggestions_ouput_path + output_name))
    
    # loads excel file and sheets
    myworkbook=openpyxl.load_workbook(str(OLS_suggestions_ouput_path + output_name))
    worksheet_info= myworkbook.get_sheet_by_name('Info')
    worksheet_no_dummies= myworkbook.get_sheet_by_name('No dummies')
    worksheet_with_dummies= myworkbook.get_sheet_by_name('With dummies')
    
    # writes information on modelled Z
    for i in range(len(statistics_dict['Dates'])):
        worksheet_info.cell(row=(i+4), column=4).value = statistics_dict['Dates'][i].strftime("%d-%m-%Y")
        worksheet_info.cell(row=(i+4), column=5).value = statistics_dict['Z'][i]
    
    # writes information on variables
    for i in range(len(statistics_dict['Variables'])):
        worksheet_info.cell(row=(i+22), column=1).value = statistics_dict['Variables'][i]
        
        # write number of transformations tried for each variable
        worksheet_info.cell(row=(i+22), column=2).value = statistics_dict['Variable transformations'][statistics_dict['Variables'][i]]
      
    # writes simulation statistisc to excel file
    worksheet_info['B3'] = statistics_dict['Combinations']
    worksheet_info['B4'] = statistics_dict['Dummy combinations']
    worksheet_info['B5'] = statistics_dict['Combinations']
    worksheet_info['B6'] = statistics_dict['With dummy combinations']
    worksheet_info['B7'] = statistics_dict['Number of simulations']
    worksheet_info['B8'] = statistics_dict['Number of no dummies']
    worksheet_info['B9'] = statistics_dict['Number with dummies']
    worksheet_info['B10'] = statistics_dict['Time to do simulations']
    
    worksheet_info['B13'] = statistics_dict['Significance level']
    worksheet_info['B14'] = statistics_dict['R level']
    worksheet_info['B15'] = statistics_dict['Max variables']
    worksheet_info['B16'] = statistics_dict['Max dummies']
    worksheet_info['B17'] = statistics_dict['Max lag']
    worksheet_info['B18'] = statistics_dict['Aim Y1']
    
    # list of models without dummies is exported
    for i in range(len(results_no_dummies)):
        
        # in order to rank the models and give them a refrence number
        worksheet_no_dummies.cell(row=(i+3), column=1).value = i + 1
        
        # writes each position in the model tuple to excel
        worksheet_no_dummies.cell(row=(i+3), column=2).value = results_no_dummies[i][0]
        worksheet_no_dummies.cell(row=(i+3), column=3).value = results_no_dummies[i][1]
        worksheet_no_dummies.cell(row=(i+3), column=4).value = results_no_dummies[i][2]
        worksheet_no_dummies.cell(row=(i+3), column=5).value = results_no_dummies[i][3]
        
        # in order to write out list of model coefficients to one cell in excel
        worksheet_no_dummies.cell(row=(i+3), column=6).value = ", ".join(list(map(lambda x: str(x), list(results_no_dummies[i][4]))))
    
    # list of models with dummies is exported
    for i in range(len(results_with_dummies)):
        
        # in order to rank the models and give them a refrence number
        worksheet_with_dummies.cell(row=(i+3), column=1).value = i + 1
        
        # writes each position in the model tuple to excel
        worksheet_with_dummies.cell(row=(i+3), column=2).value = results_with_dummies[i][0]
        worksheet_with_dummies.cell(row=(i+3), column=3).value = results_with_dummies[i][1]
        worksheet_with_dummies.cell(row=(i+3), column=4).value = results_with_dummies[i][2]
        worksheet_with_dummies.cell(row=(i+3), column=5).value = results_with_dummies[i][3]
        
        # in order to write out list of model coefficients to one cell in excel
        worksheet_with_dummies.cell(row=(i+3), column=6).value = ", ".join(list(map(lambda x: str(x), list(results_with_dummies[i][4]))))
    
    # saves and close workbook object
    myworkbook.save(str(OLS_suggestions_ouput_path + output_name))
    myworkbook.close()
        
def do_OLS(input_tuple):
    """
        Purpose: 
            Do OLS based on the specified variable combination
            This differs from the OLS function used in in-depth analysis in that this had to be optimised for execution speed. 
            For that reason it is very bare boned. 
            
        Inputs:
            input_tuple: an input tuple with the following elements:
                variables+transformations to try(tuple)
                significance level(float)
                time_series object used to fetch the time series
                r_cutoff_level(float)
                boolean value to specify if the variable tuple contains dummies or not
            
        Outputs:
            tuple with the following elements:
                variable_tuple: tuple with the explanatory variables, minus the dummy variables
                variable_info_list: tuple with all the explanatory variables labled with + or - to indicate coefficient effect
                r_squared: adjusted r squared of the model
                coefficients: the model coefficients
                variable_sign_dict: dictionary with the expected effect from the variable on the credit cycle(+ or -)
        
    """
   
    # use the time series object to get Z and variable time series
    time_series = input_tuple[2].time_series(input_tuple[0])
    Z = time_series[0]
    explanatory_variables = time_series[1] 
    
    # extract parameters from the input tuple
    significance_level = input_tuple[1]
    r_cutoff_level = input_tuple[3]
    no_dummies = input_tuple[4]
    variable_sign_dict = input_tuple[5]
    
    # prepares the variable time series for OLS regression
    explanatory_variables_array = np.array(explanatory_variables).T
    explanatory_variables_array = sm.add_constant(explanatory_variables_array)
    
    # creates OLS model
    model = sm.OLS(Z, explanatory_variables_array).fit()
    
    # correct the parameters for heteroscedasticity and autocorrelation
    model = model.get_robustcov_results(cov_type='HAC', maxlags=1)
    
    # gets adjusted r-squared and p-values
    r_squared = model.rsquared_adj
    p_values = list(model.pvalues)

    # if all variables are significant, model info is returned 
    if all(p <= significance_level for p in p_values) and r_squared >= r_cutoff_level:
        
        # gets variable combination and coefficients
        variable_combination = input_tuple[0]
        coefficients = model.params
        coefficients = coefficients.tolist()
        
        # calculates mean squared error
        MSE = statistics.mean(list(map(lambda x: x**2, list(model.resid))))
                
        variable_info_list = []
        expected_sign = True
        # iterates through all explanatory variables to check that the value of the variable coefficient has the expected sign(+/-)
        for i in range(len(variable_combination)):
            
            # if coefficient is bellow zero, check if this corresponds with expected sign
            if coefficients[1 + i] < 0:
                variable_name = variable_combination[i].split()[0]
                if "Dummy" not in variable_name:
                    if variable_sign_dict[variable_name] != "-":
                        expected_sign = False
                        break
                    
                variable_info_list.append("(-)" + variable_combination[i])
                    
            # if coefficient is above zero, check if this corresponds with expected sign
            else:
                variable_name = variable_combination[i].split()[0]
                if "Dummy" not in variable_name:
                    if variable_sign_dict[variable_name] != "+":
                        expected_sign = False
                        break
                        
                variable_info_list.append("(+)" + variable_combination[i])

        # discards models with wrong coefficient signs
        if expected_sign:
        
            # in order to remove the dummies names to help analysts find optimal model
            if no_dummies:                
                return (input_tuple[0], variable_info_list, r_squared, MSE, coefficients)
            else:
                variables_without_dummies = []
                for variable in input_tuple[0]:
                    if "Dummy" not in variable:
                        variables_without_dummies.append(variable)

                variable_tuple = tuple(variables_without_dummies)
                return (variable_tuple, variable_info_list, r_squared, MSE, coefficients)

def combine_lists(input_list):
    """
        Purpose: 
            Takes n number of lists as input and combines the list elements to generate a list of combinations
            
        Inputs:
            input_list: a list of the lists to be combined                        
            
        Outputs:
            a list of all combinations of the elements of the inputted lists
        
    """
    return_list = []
    
    for combination in itertools.product(*input_list):      
        # join all tuples to one tuple and appends to list to be returned
        return_list.append(sum(combination, ()))
      
    return return_list


def combine(variables, max_combinations, variable_name=None):
    """
        Purpose: 
            Generate all possible combinations of the inputted variables
            
        Inputs:
            variables: a list or dictionary containing the variables to be combined
            max_combinations: maximum number of variables in one combination
            variable_name: an optional parameter specifying the variable name
            
        Outputs:
            a list of all combinations of the elements of the inputted lists, where each element are tuples
        
    """

    number_of_variables = len(variables)
    
    # if input is a dictionary, this is transformed to a list
    if isinstance(variables, dict):
        variable_name_list = list(variables.keys())
    else:
        variable_name_list = variables
    
    # adds the variable name if specified
    if variable_name:
        variable_name_list = list(map(lambda x: variable_name + " " + x, variable_name_list))
    
    # transforms the elements in variable_name_list into tuples
    variable_name_list = list(map(lambda x: tuple([x]), variable_name_list))
    
    # adds the variables without combinations to the return list as well
    return_list = variable_name_list
    
    # checks if there are any combinations to be done
    if max_combinations > 1:
        if number_of_variables < max_combinations:
            max_combinations = number_of_variables
        
        # combines the variables
        combinations_list = []
        for length in range(2, max_combinations + 1):
            for combination in itertools.combinations(variable_name_list, length):
                combinations_list.append(tuple(list(combination)))

        return_list.extend(combinations_list)    
    
    return return_list


def do_variable_transformations(variable_name, macro_variable_actuals, macro_variable_projections, do_ln=True, inflation_data_actuals=None,
                                inflation_data_projections=None, inflation_type=None, inflation_metadata=None, max_lag=3):
    """
               Purpose:
               Transform a variable into all specified transformation types

               Inputs:
                macro_variable: dictionary with key=date and value=macro variable value
                do_ln: boolean variable specifying if ln transformation should be done
                inflation_data: dictionary with key=date and value=inflation value
                inflation_type: string variable specifying if US or Norwegian inflation data should be used.
                                If None, no inflation adjustment will be done.
                inflation_metadata: dataframe with information of start data for inflation adjustment for US and Norway

               Outputs:
                Dictionary with key=transformation type, value = dictionary with key=date, value=transformed value

    """
    # populates the list of ln and inflation transformations
    additional_transformation_list = [""]
    
    # sets transformation parameters
    if do_ln:
        additional_transformation_list.append("Ln")

    if inflation_type:
        if inflation_type.upper() == "NOR":
            additional_transformation_list.append("RealNOR")
        elif inflation_type.upper() == "US":
            additional_transformation_list.append("RealUS")

        if do_ln and inflation_type.upper()=="NOR":
            additional_transformation_list.append("LnRealNOR")
        elif do_ln and inflation_type.upper()=="US":
            additional_transformation_list.append("LnRealUS")
 
    # the base transformations to be done
    transformation_list = ["",                             
                           "Pct1", "Pct2", "Pct3",
                           "Delta1", "Delta2", "Delta3", 
                           "Maxdelta2", "Maxdelta3",
                           "Movavg2",  
                           "Pct1Movavg2", 
                             ]
    
    # in order to be able to controll the lags of the model. The more lagged variables the less responsive the model is to financial shocks
    if max_lag >= 1:
        transformation_list.extend(["Lag1", "Movavg3", "Pct1Movavg3", "Lag1Pct1", "Lag1Pct2", "Lag1Delta1", "Lag1Delta2"])
        
        if max_lag >= 2:
            transformation_list.extend(["Lag2", "Movavg4", "Pct1Movavg4", "Lag2Pct1", "Lag2Pct2", "Lag2Delta1", "Lag2Delta2"])
                
            if max_lag >= 3:
                transformation_list.extend(["Lag3", "Lag3Pct1", "Lag3Pct2", "Lag3Delta1", "Lag3Delta2"])
                           

    transformation_dict = {}
    transformation_object = Dataprep_Functions.Ols_parameter_transformation()
    
    # if ln and/or inflation adjustment is to be done, this is looped over
    for adjust_variable in additional_transformation_list:
        
        # loops over all the base transformations to be done
        for transformation_type in transformation_list:
            transform_type = adjust_variable + transformation_type                      
            
            # tries to do transformation. As some values can't have all transformations done, a try/except must be used to avoid crashing
            try:
                transformation_dict[transform_type] = \
                transformation_object.transform_variable(
                    actuals=macro_variable_actuals, projections=macro_variable_projections, transform_type=transform_type,
                    US_actual_inflation=inflation_data_actuals, NOR_actual_inflation=inflation_data_actuals,
                    US_projected_inflation=inflation_data_projections, NOR_projected_inflation=inflation_data_projections,
                    inflation_metadata=inflation_metadata)
            except:
                print(transform_type, "could not be done on variable", variable_name)
                                
    
    return transformation_dict


def generate_dummies(date_list, max_dummy_length=5):
    """
        Purpose: 
            Generate dummmies based on the inputted date list
            
        Inputs:
            date_list: a list of the dates in the Z time series                       
            max_dummy_length: maximum length of the dummies
            
        Outputs:
            a dictionary of all dummies time series given the inputted date list with key=year(s) value=dummy time series
        
    """
    
    # to avoid the program trying to make dummy time series longer than the Z time series
    time_series_length = len(date_list)
    if max_dummy_length > time_series_length:
        max_dummy_length = time_series_length -1
        
    dummy_dict = {}

    # iterate over all dummy lengths
    for i in range(1, max_dummy_length+1):

        # iterate over all start dates
        for j in range(time_series_length-i+1):

            length_counter = i
            start_date = None
            end_date = None
            dummy_series = []

            # iterate over all years in the time series
            for x in range(1, time_series_length+1):

                # appends 1 if year is higher than start date for dummie, and the length counter hasn't reached zero
                if x > j and length_counter > 0:
                    dummy_series.append(1)

                    if start_date == None:
                        start_date = date_list[x - 1]

                    end_date = date_list[x-1]
                    length_counter -= 1
                    
                # append 0 before and after dummy period
                else:
                    dummy_series.append(0)
            
            # in order to write out the correct dummy name
            if start_date == end_date:
                dummy_name = "Dummy " + start_date.strftime("%Y")
            else:
                dummy_name = "Dummy " + start_date.strftime("%Y") + "-" + end_date.strftime("%Y")

            dummy_dict[dummy_name] = dummy_series

    return dummy_dict


class TimeSeriesLookup:
    '''
    A class to store all the time series used in the estimation
    
    '''
    
    # constructor to create object of class TimeSeriesLookup with time series to use in OLS regression
    def __init__(self, variables_dict, dummy_dict, Z):

        for variable in variables_dict:
            dict_values = variables_dict[variable].values()
            variables_dict[variable] = list(list(zip(*dict_values))[0])

        # creates one dictionary from variables_dict and dummy_dict
        self.time_series_dict = {**variables_dict, **dummy_dict}

        self.Z = Z

    def return_all_time_series(self):
        return (self.Z, self.time_series_dict)

    # returns the time series for Z and for the input variables
    # trims all the time series to match the length of the shortest one
    def time_series(self, ex_variables_tuple):
        return_Z = self.Z
        time_series_list = []
        lowest_length = 10000
        for variable in ex_variables_tuple:
            
            time_series_list.append(self.time_series_dict[variable])            
            
            #find time series with lowest length
            if len(self.time_series_dict[variable]) < lowest_length:
                lowest_length = len(self.time_series_dict[variable])
        
        #trim variable times series in order to have them same length
        for i in range(len(time_series_list)):
            if len(time_series_list[i]) > lowest_length:
                time_series_list[i] = time_series_list[i][len(time_series_list[i])-lowest_length:]
        
        #trim Z time series in order to have it at same length as variables
        if len(return_Z) > lowest_length:
            return_Z = self.Z[len(self.Z)-lowest_length:]
        
        return return_Z, time_series_list
