import pandas as pd
import os
import xlrd
import shutil

import sys
sys.path.append("../")
import Macro_Model_Tool

# the user interface class
class User:
    
    def __init__(self, macro_project):
        '''
        Constructor of the class User.
        Gives a way to keep chosen user variables stored. 
        The most significant information stored in the object is the default_path_dict, which allows the code to interact witht he whole project
        '''
        
        # gets the default paths used in the ECL project, and expands it with the paths required in the macro project
        #self.file_path_pd = pd.DataFrame()
        self.default_path_dict = get_default_file_paths()
        self.project_path = str("../../../data/Macro_estimation_projects/" + macro_project)
        self.default_path_dict["Macro_list_parameters"] = str(self.project_path + "/Project_parameters/Simulation_input.xlsx")
        self.default_path_dict["Macro_variables"] = str(self.project_path + "/Project_parameters/Macro_data.xlsx")
        self.default_path_dict["Z_values"] = str(self.project_path + "/Project_parameters/Z_values.xlsx")
        self.default_path_dict["OLS_output_template"] = str(self.project_path + "/Project_parameters/OLS_results_template.xlsx")
        self.default_path_dict["OLS_suggestions_template"] = str(self.project_path + "/Project_parameters/OLS_suggestions_template.xlsx")
        self.default_path_dict["OLS_output_path"] = str(self.project_path + "/OLS_analysis_results/")
        self.default_path_dict["OLS_simulation_output_path"] = str(self.project_path + "/Suggested_OLS_models/")
        self.default_path_dict["Project_coefficients_path"] = str(self.project_path + "/Project_model_coefficients/")
        
        macro_sheet_df = pd.read_excel(str(self.project_path + "/Project_parameters/Macro_sheets.xlsx"))
        macro_sheet_dict = dict(zip(macro_sheet_df['Type'], macro_sheet_df['Name']))
        self.default_path_dict["Macro_sheet_dict"] = macro_sheet_dict
        

        self.action_dict = {
            '1': "Generate Credit Cycles",
            '2': "Generate model suggestions",
            '3': "Do in-depth analysis of specified model",
            '4': "Validate existing models",
            '5': "Create new model parameters"
        }  
        
        self.macro_project_name = macro_project    
        
        
    def set_purpose(self):
        """
        Purpose:
        Interface to allow user to specify what their purpous is
        """
        long_break_line()
        print("Working on project:", self.macro_project_name)
        short_break_line()
        print("What is your purpose with the Macro Model Tool today?")

        user_action = 0
        for action in self.action_dict:
            print(self.action_dict[action], ", type", action)
        
        short_break_line()
        while user_action not in self.action_dict:
            user_action = input("Your purpose: ")

        self.purpose = user_action
        
    def fire_it_up(self):
        """
              Purpose:
              Starts the correct flow based on what purpose the user has
              """
        action_function_dict = {
            '1': lambda: self.generate_credit_cycles(),
            '2': lambda: self.generate_model_suggestions(),
            '3': lambda: self.in_depth_analysis(),
            '4': lambda: self.validate_existing(),
            '5': lambda: self.deploy_new_model_parameters()            
        }

        action_function_dict.get(self.purpose)()
        

        
    def generate_credit_cycles(self):
        print("This functionality is not currently implemented")
    
    

    def generate_model_suggestions(self):
        """
        Purpose:
        An interface to get user input required to call generate_OLS_models() in Macro_Model_Tool.py
        """
        
        # promts the user to specify the macro model
        macro_model = self.specify_model()
        
        short_break_line()
        
        # allows the user to specify variables to include in OLS suggestion simulation
        variables_df = self.set_simulation_variables(macro_model)
        
        short_break_line()
        
        # allows user to specify maximum number of lag years
        user_response = ""
        while user_response not in ["0", "1", "2", "3"]:
            user_response = input("Maximum number of lag years to use(0-3): ")
        
        try:
            max_lag = int(user_response)
        except:
            max_lag = ""
        
        short_break_line()
        
        # allows user to specify if there should be a percentile aim for first year of predictions
        user_response = ""
        while user_response != "Y" and user_response != "N":
            user_response = input("Do you want to set a target percentile for first year of predictions(Y/N): ").upper()
        
        # if aim is to set, this allows user to specify the aim of year 1
        if user_response == "Y":
            target_Z_percentile = -1
            while target_Z_percentile <= 0 or target_Z_percentile >= 1:
                target_Z_percentile = input("Specify what Z percentile to aim for first year of predictions(0-1): ")
                
                # converts from Norwegian decimal format to English format
                target_Z_percentile = target_Z_percentile.replace(",", ".")
                
                try:
                    target_Z_percentile = float(target_Z_percentile)
                except:
                    target_Z_percentile = -1
        else:
            target_Z_percentile = None
        
        short_break_line()
        
        # allows the user to use default parameters
        print("Default parameters for OLS suggestions are specified in the user guide.")
        print("Be aware that execution time might increase considerably if including more variables, dummies and transformations of the same variables than the default")
        user_response = ""
        while user_response != "Y" and user_response != "N":            
            user_response = input("Do you want to use default parameters for OLS regression(Y/N): ").upper()
        
        short_break_line()
        # if not default parameters, the user can specify the parameters
        if user_response == "N":
            
            # allows the user to specify maximum number af variables to include in the simulation
            max_variables=0
            while max_variables < 1 or max_variables > 4:
                max_variables = input("What's the maximum number of variables (1-4): ")
                try:
                    max_variables = int(max_variables)
                except:
                    max_variables = 0
            
            short_break_line()
            # allows the user to specify maximum number af dummies to include in the simulation 
            max_dummies=0
            while max_dummies < 1 or max_dummies > 3:
                max_dummies = input("What's the maximum number of dummies (1-3): ")
                try:
                    max_dummies = int(max_dummies)
                except:
                    max_dummies = 0
            
            short_break_line()
            # allows the user to specify maximum number transformations of the same variable to include
            max_one_variable=0
            while max_one_variable < 1 or max_one_variable > 2:
                max_one_variable = input("What's the maximum number of transformations of the same variable(1-2): ")
                try:
                    max_one_variable = int(max_one_variable)
                except:
                    max_one_variable = 0
            
            short_break_line()
            # allows the user to specify the significance level to be used as a cuttof value in the the simualtion
            significance_level=0
            while significance_level <= 0 or significance_level >= 1:
                significance_level = input("Specify significance level for variables(between 0 and 1): ")
                significance_level = significance_level.replace(",", ".")
                try:
                    significance_level = float(significance_level)
                except:
                    significance_level = 0
            
            short_break_line()
            # allows the user to specify the r-squared level to be used as a cuttof value in the the simualtion
            r_cutoff_level=0
            while r_cutoff_level <= 0 or r_cutoff_level >= 1:
                r_cutoff_level = input("Specify r squared cutoff value(between 0 and 1): ")
                r_cutoff_level = r_cutoff_level.replace(",", ".")
                try:
                    r_cutoff_level = float(r_cutoff_level)
                except:
                    r_cutoff_level = 0
            short_break_line()
            
            # calls the function with specified OLS parameters
            Macro_Model_Tool.generate_OLS_models(path_dict=self.default_path_dict, macro_model=macro_model, variables=variables_df, max_variables=max_variables, max_dummies=max_dummies, max_one_variable=max_one_variable, significance_level=significance_level, r_cutoff_level=r_cutoff_level, max_lag=max_lag, Y1_predict_aim=target_Z_percentile)
            
        # calls the function with default parameters if default parameters is used
        else:            
            Macro_Model_Tool.generate_OLS_models(path_dict=self.default_path_dict, macro_model=macro_model, variables=variables_df, max_lag=max_lag, Y1_predict_aim=target_Z_percentile)
        
        short_break_line()
        print("Done with simulations")
        print("Results are saved in the project folder under Suggested_OLS_models")
        
        
    
    def in_depth_analysis(self):
        """
        Purpose:
        An interface to get user input required to call in_depth_analysis() in Macro_Model_Tool.py
        """
        
        # promts the user to specify the macro model
        macro_model = self.specify_model()
        
        # gives users the possibility to choose what way to declare variables and transformations to the analysis
        # only one possibility as of now, but functionality is made such that this can be expanded on in the future
        short_break_line()
        print("Variable and transformations input options")
        input_options = {"1": "Type/paste in variables and transformations"}
        for input_option in input_options:
            print(input_option, ":", input_options[input_option])
        
        short_break_line()
        user_input = ""
        while user_input not in input_options.keys():
            user_input = input("Select the number corresponding with the input option you want to use: ").upper()
        
        # gives the user the input to type or paste in variables and transformation to be used in analysis
        short_break_line()
        if user_input == "1":
            print("Format of variable input: variableName1 Transformation(if any), variableName2 Transformation(if any),....")
            print("Remember to capitalize first letter in transformation")
            
            user_input_variables = input("Type/paste your variables here: ")
                
            # have a try clause that will handle incorrect format of input
            try:
                if "(+)" in user_input_variables or "(-)" in user_input_variables:
                    
                    # removes extra characters that might be included when pasting in variables and transformations
                    user_input_variables = user_input_variables.replace("(+)", "").replace("(-)", "").replace("'", "")

                variables = []
                # splits the variables with transformations based on comma
                user_input_variables = user_input_variables.split(",")
                for variable in user_input_variables:
                    variable = variable.strip()
                    
                    # dummy variables are converted to format required by program
                    if "Dummy" in variable:
                        variable_name = variable
                        variable_transformation = ""
                    else:                        
                        # splits on space to seperate variable and transformation
                        variable_split = variable.split(" ")
                        
                        variable_name = variable_split[0]                        
                        if len(variable_split) == 2:
                            variable_transformation = variable_split[1]
                            
                        # if variable_split only has length of one, it means no transformation will be done
                        else:
                            variable_transformation = ""
                    
                    # converts the set of variable and transformation into a tuple to fit the format in the rest of the program
                    variable_tuple = (variable_name, variable_transformation)
                    variables.append(variable_tuple)
                    
            # if the above code fails, it is assumed the user did not input the variables and transformation in the correct format
            except:
                print("Pasted parameters was not on correct format")
                user_input = ""
        
        # have a try except clause to avoid crash if given variables or transformations aren't supported in program, or if unexpected errors happens
        try:
            OLS_results = Macro_Model_Tool.in_depth_analysis(macro_model_name=macro_model, explanatory_variables=variables, path_dict=self.default_path_dict, rolling_years_amount=10)
            
            # in case execution was aborted due to error in macro input
            if OLS_results:
                # extract OLS parameters from OLS_results object
                variables = OLS_results.get_variables()
                coefficients = OLS_results.get_coefficients()
                p_values = OLS_results.get_p_values()

                # to map variables with transformations, coefficients and p-values
                variable_info_list = [["Intercept", "", p_values[0], coefficients[0]]]
                for i in range(len(variables)):
                    variable_info_list.append(list([variables[i][0], variables[i][1], p_values[i+1], coefficients[i+1]]))

                # prints general model results
                short_break_line()
                print("Model results for", OLS_results.get_model_name())
                print("R-squared: ", round(OLS_results.get_adjusted_r_squared(), 6))
                variables_info_df = pd.DataFrame(variable_info_list, columns=["Variable", "Transformation", "P-values", "Coefficient"])
                print(variables_info_df.round(6))

                short_break_line()

                # allows the user to save model results for later deployment to ECL model
                user_input = ""
                while user_input != "Y" and user_input != "N":
                    user_input = input("Do you want to save model parameters for later deployment to the ECL model(Y/N): ").upper()

                    if user_input == "Y":
                        user_input = input("Confirm by writing 'Save': ").replace("'", "").replace('"', "").upper()
                        if user_input == "SAVE":
                            Macro_Model_Tool.save_model_results(OLS_results.get_model_name(), variables_info_df, self.default_path_dict)
                            short_break_line()
                            print("Model parameters saved")
                            user_input = "Y"
                        else:
                            user_input = ""
        
        # the program makes a best guess as to why the program crashed, and gives the users information as to how this can be solved         
        except KeyError:
            short_break_line()
            print("Oh no, something went wrong!")
            print("Seems like at least one of the variables you stated isn't supported in this program")
            print("Please make sure all variables exist in the macro input sheets")
            print("Note that some variables contain capitalized letters, and must be typed that way. The variable Dummy must have a capitalized D")
            print("If all variables do exist in all macro sheets, contact model owner")
         
        except AssertionError:
            short_break_line()
            print("Oh no, something went wrong!")
            print("Seems like at least one of the transformation you stated isn't supported in this program")
            print("Please make sure the transformation exist in this program's user guide and that the first letter of transformation names are Capitalized")
            print("If all transformations exist in the user guide, and the first letter of the transformations are capitalized, contact model owner")
        
        
        #except Exception as e:
         #   short_break_line()
          #  print("Oh no, something unexpected went wrong!")
           # print("Contact model owner if this error persists")
            #print(e)
            
        
        
        
    def validate_existing(self):
        print("Functionality not yet implemented")
    
    def deploy_new_model_parameters(self):
        """
        Purpose:
        An interface that allows user to deploy new model parameter to the ECL model
        """
        
        changed_models_df = pd.read_excel(str(self.default_path_dict["Project_coefficients_path"] + "Changed_models.xlsx"), header=0)
        
        short_break_line()
        # iterates the df to print out the changed models with change date to the user
        print("These are the updated models with last update date")
        for row in changed_models_df.itertuples():
            #print(row)
            if row.Changed == "Yes":
                print(row.Model, row.Date)
        
        user_input = ""
        while user_input != "Y" and user_input != "N":
            user_input = input("Do you want to deploy new model coefficients to ECL model(Y/N): ").upper()
            
            if user_input == "Y":
                # includes an extra step to minimize the possibility of deploying the variables by accident
                user_input = input("Confirm by typing 'Deploy': ").replace("'", "").replace('"', "").upper()
                if user_input == "DEPLOY":
                    short_break_line()
                    Macro_Model_Tool.deploy_new_models(self.default_path_dict, self.macro_project_name )
                    user_input = "Y"
                else:
                    user_input = ""    
                    
   
    def specify_model(self):
        """
        Purpose:
        Allows for the user to specify model to work on
        """
        
        # gets the list of IFRS9 macro models used in the ECL program
        segment_info = pd.read_excel(self.default_path_dict["Segment_info"], sheet_name="Info")
        macro_models = list(segment_info["MACRO_MODEL"])
        
        short_break_line()
        # for removing duplicates and sorting
        macro_models = list(set(macro_models))
        macro_models.sort()
        
        # prints the IFRS9 macro models used in the ECL program and allows the user to select one
        macro_model_dict = {}
        counter = 1
        print("Macro models:")
        for model in macro_models:
            print(counter, ":", model)
            macro_model_dict[counter] = model   
            counter += 1
        
        short_break_line()
        chosen_model = 0
        while chosen_model not in macro_model_dict:            
            chosen_model = input("Type the number corresponding to the macro model you want to work on: ")
            try:
                chosen_model = int(chosen_model)
            except:
                chosen_model = 0
            
        return macro_model_dict[chosen_model]
    
    def set_simulation_variables(self, model_name):
        """
        Purpose:
        Allows for the user to specify variables to be included in the OLS suggestions simulation
        """
        
        variable_df = pd.read_excel(self.default_path_dict["Macro_list_parameters"], sheet_name=model_name)     
        
        # selection of variables from the user
        done_selecting_variables = False
        while not done_selecting_variables:
            
            #prints the possible variables for the user
            variable_dict = {}
            counter = 1
            print("Variables for OLS simulation:")
            for index, row in variable_df.iterrows():
                variable_dict[counter] = row[0]
                print(counter, " : ", row[0], ", expected sign:", row[3])
                counter += 1
            
            # allows the user to specify variables to go into the OLS suggestion simulation, and warns of the increase in execution time if too many variables are included
            short_break_line()
            print("Be aware that execution time will increase considerably for each extra variable added to the simulation")
            print("The recommondation is to use a maximum of 4 input variables, as 5 or more variables might lead to impractical execution times of more than 15 minutes")
            print("This also depends heavily on number of variables transformation to be used.")
            short_break_line()
            
            user_selection = []
            # checks that all variables selected are valid
            while len(user_selection) < 1 or not all(item in variable_dict.keys() for item in user_selection):
                user_selection = input("Input the numbers corresponding to the variables you want to use in the simulation. Separate the values with ',': ").split(",")

                # tries to convert to int
                try:
                    user_selection = list(map(lambda x: int(x), user_selection))
                except:
                    user_selection = []

            short_break_line()   
            # prints out the variables for the user to controll that the expected variables were selected
            print("Variables selected for simulation:")
            for selection in user_selection:
                print("- ", variable_dict[selection])
            
            short_break_line()
            # allows for the user to select variables again if something went wrong
            user_input = ""
            while user_input != "Y" and user_input != "N":
                user_input = input("Done selecting variables(Y/N): ").upper()
            
            if user_input == "Y":
                done_selecting_variables = True
            else:
                done_selecting_variables = False
        
        # filters away non-selected variables
        variables = list(map(lambda x: variable_dict[x], user_selection))        
        return_variable_df = variable_df[variable_df['Variables'].isin(variables)]
        
        return return_variable_df
        
        
def get_default_file_paths():
    """
            Purpose:
            Fetches the default paths from the directory structure.
            Puts it in a data frame, and creates a dict of that dataframe

    """
    default_paths_path = "../../../data/ECL_Data/File_mapping/Current_default_paths.xlsx"
    file_path_pd = pd.read_excel(default_paths_path)

    path_dict = {}

    for index, row in file_path_pd.iterrows():
        path_dict[row[0]] = str(row[1])

    return path_dict
    
def long_break_line():
    print("--------------------------------------------------------------------")

def short_break_line():
    print("---------------------------------")

        
def create_macro_project(projects):
    """
            Purpose:
            Give the user the possibility to create a nwe macro estimation project with the specified input files

    """
    
    long_break_line()
    print("Before making a new project, make sure this data is uploaded on IPA under Macro_tool_data:")
    print(" - Macro data with baseline and adverse")
    print(" - Z values to be modelled")
    print(" - List of input variables to be used for suggestion module")
    print("Description of how to upload and the expected format is described in the ECL Macro Tool User Guide")
    
    long_break_line()
    
    # gets new project name and checks that a project with the same name doesn't exist
    project_name = ""
    while project_name == "" or project_name.upper() in list(map(lambda x: x.upper(), projects)):
        project_name = input("Name of new macro project: ")
        
        if project_name.upper() in list(map(lambda x: x.upper(), projects)):
            print("Project name is taken. Please create a unique project name")
            short_break_line()
    
    # sets the macro data for the project
    macro_files = {}
    counter = 1
    macro_path = str("../../../data/Macro_tool_data/Macro_data")
    for file in os.listdir(macro_path): 
        # in order to not show hidden files
        if file[0] != ".":            
            macro_files[counter] = file
            counter += 1
        
    print("Macro forecast files:")
    for file in macro_files:
        print(file, ":", macro_files[file])
        
    short_break_line()
    macro_data_file = 0
    baseline_sheet = ""
    adverse_sheet = ""
    while macro_data_file not in macro_files.keys():
        macro_data_file = input("Please select number corresponding to the correct macro data for project: ")
        try:
            # sets the baseline, low and adverse macro data for the project
            macro_data_file = int(macro_data_file)
            macro_file = xlrd.open_workbook(macro_path + "/" + macro_files[macro_data_file], on_demand=True)
            macro_file_sheets = macro_file.sheet_names()
            
            sheet_dict = {}
            counter = 1
            for sheet in macro_file_sheets:
                sheet_dict[counter] = sheet
                counter += 1

            short_break_line()
            print("Macro forecast sheets:")
            for sheet in sheet_dict:
                print(sheet, ":", sheet_dict[sheet])
            
            baseline_sheet_number = 0
            adverse_sheet_number = 0
            low_sheet_number = 0
            
            # the user must select three unique sheets for different macro scenarions from the sheets in the given macro file
            while adverse_sheet_number not in sheet_dict.keys() or baseline_sheet_number not in sheet_dict.keys() or low_sheet_number not in sheet_dict.keys() or baseline_sheet_number == adverse_sheet_number or baseline_sheet_number == low_sheet_number or low_sheet_number == adverse_sheet_number:
                short_break_line()
                baseline_sheet_number = input("Please select the number corresponding to the sheet for baseline forecasts: ")
                adverse_sheet_number = input("Please select the number corresponding to the sheet for adverse forecasts: ")
                low_sheet_number = input("Please select the number corresponding to the sheet for low forecasts: ")
                try: 
                    baseline_sheet_number = int(baseline_sheet_number)
                    adverse_sheet_number = int(adverse_sheet_number)
                    low_sheet_number = int(low_sheet_number)
                except:
                    baseline_sheet_number = 0
                    adverse_sheet_number = 0
                    low_sheet_number = 0
          
            # add code to validate the macro input
        except:
            macro_data_file = 0
          
    # sets the Z values for the project
    z_files = {}
    counter = 1
    z_path = str("../../../data/Macro_tool_data/Segment_Z")
    for file in os.listdir(z_path): 
        if file[0] != ".":
            z_files[counter] = file
            counter += 1
    
    # sets the Z to be modelled in the project
    short_break_line()
    print("Segment Z files:")
    for file in z_files:
        print(file, ":", z_files[file])
    
    short_break_line()
    z_file = 0
    while z_file not in z_files.keys():
        z_file = input("Please select the number corresponding to the Z values to be used in the project: ")
        try:
            z_file = int(z_file)
        except:
            z_file = 0
          
    # sets the input for OLS simulations for the project
    parameter_files = {}
    counter = 1
    parameter_path = str("../../../data/Macro_tool_data/OLS_simulation_input")
    for file in os.listdir(parameter_path): 
        if file[0] != ".":
            parameter_files[counter] = file
            counter += 1
            
    short_break_line()    
    print("OLS simulation input files:")
    for file in parameter_files:
        print(file, ":", parameter_files[file])
    
    short_break_line()
    parameter_file = 0
    while parameter_file not in parameter_files.keys():
        parameter_file = input("Please select the number corresponding to the variable input to the suggestion module to be used in the project: ")
        try:
            parameter_file = int(parameter_file)
        except:
            parameter_file = 0
    
    # create project folder and subfolders
    project_path = "../../../data/Macro_estimation_projects/" + project_name
    os.mkdir(project_path)
    os.mkdir(project_path + "/OLS_analysis_results")
    os.mkdir(project_path + "/Suggested_OLS_models")
    os.mkdir(project_path + "/Project_parameters")    
    os.mkdir(project_path + "/Project_model_coefficients")

          
    # copy macro data
    shutil.copy(str(macro_path + "/" + macro_files[macro_data_file]), str(project_path + "/Project_parameters/Macro_data.xlsx"))
    # copy Z values
    shutil.copy(str(z_path + "/" + z_files[z_file]), str(project_path + "/Project_parameters/Z_values.xlsx"))
    # copy input to OLS simulation
    shutil.copy(str(parameter_path + "/" + parameter_files[parameter_file]), str(project_path + "/Project_parameters/Simulation_input.xlsx"))
    # copy template for OLS analysis output
    shutil.copy(str("../../../data/Macro_tool_data/Analysis_output_template/OLS_results_template.xlsx"), str(project_path + "/Project_parameters/OLS_results_template.xlsx"))
    # copy template for OLS suggestions output
    shutil.copy(str("../../../data/Macro_tool_data/Suggestion_output_template/OLS_suggestions_template.xlsx"), str(project_path + "/Project_parameters/OLS_suggestions_template.xlsx"))
    
    # saves mapping to macro data sheets
    project_macro_sheets = [["Baseline_sheet", sheet_dict[baseline_sheet_number]], ["Adverse_sheet", sheet_dict[adverse_sheet_number]], ["Low_sheet", sheet_dict[low_sheet_number]]]
    macro_sheets = pd.DataFrame(project_macro_sheets, columns=["Type", "Name"])
    macro_sheets.to_excel(project_path + "/Project_parameters/Macro_sheets.xlsx")
    
    # copies current macro models to project
    default_path_dict = get_default_file_paths()
    shutil.copy(str(default_path_dict['Macro_models']), str(project_path + "/Project_model_coefficients/Macro_models.xlsx"))
    
    # copies current IFRS9 macro model parameters, and create file keeping track of models changed in this particular project
    macro_models = xlrd.open_workbook(str(project_path + "/Project_model_coefficients/Macro_models.xlsx"), on_demand=True).sheet_names()        
    changed_models = pd.DataFrame(data=list(map(lambda x: list([x, "No", ""]), macro_models)), columns=['Model', 'Changed', 'Date']) 
    changed_models.to_excel(str(project_path + "/Project_model_coefficients/Changed_models.xlsx"))
    
    short_break_line()
    print("Project", project_name, "created")
    print("Navigating to", project_name)
          
    return project_name
    
        
def set_macro_project():
    """
            Purpose:
            Allows the user to choose what macro estimation project to work on
    """
    
    macro_projects = {}
    counter = 1    
    macro_projects_path = str("../../../data/Macro_estimation_projects")
    for file in os.listdir(macro_projects_path): 
        if file[0] != ".":
            macro_projects[counter] = file
            counter += 1
            
        
    print("Macro model estimation projects:")
    for file in macro_projects:
        print(file, ":", macro_projects[file])
    
    short_break_line()
    macro_project = 0
    while macro_project not in macro_projects.keys() and macro_project != "NEW":
        macro_project = input("Please select number corresponding to the correct macro project. For new project type 'new': ")
        
        try:
            macro_project = int(macro_project)
        except:
            try:
                macro_project = macro_project.replace("'", "").upper().strip()
            except:
                macro_project = 0
    
    # calls function to create new project if user type NEW
    if macro_project == "NEW":
        macro_project_return = create_macro_project(list(macro_projects.values()))
    else:    
        macro_project_return = macro_projects[macro_project]
        
    
    return macro_project_return
        
    
                     
def main():    
    """
            Purpose:
            This function initiates the entire program

    """
    while True:
        
        # use a try-except to handle unexpected exceptions in the program
        # the commented out if-else is for testing and recreating error stack. 
        #if 1==1:
        try:        
            long_break_line()
            print("Welcome to the IFRS9 Macro Model Tool")
            short_break_line()
            macro_project = set_macro_project()

            while True:
                user_object = User(macro_project)
                user_object.set_purpose()

                user_object.fire_it_up()
        #else:
        except Exception as e: 
            print("An unexpected error has occured")
            print("Please contact model owner is this error persist and attach this error message:")
            print(e)

if __name__ == "__main__":
    main()